# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import inspect
import re

from oslo_config import cfg
from oslo_log import log as logging
import six
from stevedore import dispatch

from murano.common.i18n import _LE, _LI, _LW
from murano.dsl import murano_package


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

# regexp validator to ensure that the entry-point name is a valid MuranoPL
# class name with an optional namespace name
NAME_RE = re.compile(r'^[a-zA-Z]\w*(\.[a-zA-Z]\w*)*$')


class PluginLoader(object):
    def __init__(self, namespace="io.murano.extensions"):
        LOG.info('Loading extension plugins')
        self.namespace = namespace
        extension_manager = dispatch.EnabledExtensionManager(
            self.namespace,
            PluginLoader.is_plugin_enabled,
            on_load_failure_callback=PluginLoader._on_load_failure)
        self.packages = {}
        name_map = {}
        for ext in extension_manager.extensions:
            self.load_extension(ext, name_map)
        self.cleanup_duplicates(name_map)

    def load_extension(self, extension, name_map):
        dist_name = str(extension.entry_point.dist)
        name = extension.entry_point.name
        if not NAME_RE.match(name):
            LOG.warning(_LW("Entry-point 'name' {name} is invalid").format(
                name=name))
            return
        name_map.setdefault(name, []).append(dist_name)
        if dist_name in self.packages:
            package = self.packages[dist_name]
        else:
            package = PackageDefinition(extension.entry_point.dist)
            self.packages[dist_name] = package

        plugin = extension.plugin
        try:
            package.classes[name] = initialize_plugin(plugin)
        except Exception:
            LOG.exception(_LE("Unable to initialize plugin for {name}").format(
                name=name))
            return
        LOG.info(_LI("Loaded class {class_name} from {dist}").format(
                 class_name=name, dist=dist_name))

    def cleanup_duplicates(self, name_map):
        for class_name, package_names in six.iteritems(name_map):
            if len(package_names) >= 2:
                LOG.warning(_LW("Class is defined in multiple packages!"))
                for package_name in package_names:
                    LOG.warning(_LW("Disabling class {class_name} in {dist} "
                                    "due to conflict").format(
                                        class_name=class_name,
                                        dist=package_name))
                    self.packages[package_name].classes.pop(class_name)

    @staticmethod
    def is_plugin_enabled(extension):
        if CONF.murano.enabled_plugins is None:
            # assume all plugins are enabled until manually specified otherwise
            return True
        else:
            return (extension.entry_point.dist.project_name in
                    CONF.murano.enabled_plugins)

    @staticmethod
    def _on_load_failure(manager, ep, exc):
        LOG.warning(_LW("Error loading entry-point {ep} from package {dist}: "
                        "{err}").format(ep=ep.name, dist=ep.dist, err=exc))

    def register_in_loader(self, package_loader):
        for package in six.itervalues(self.packages):
            package_loader.register_package(
                MuranoPackage(package_loader, package))


def initialize_plugin(plugin):
    if hasattr(plugin, "init_plugin"):
        initializer = getattr(plugin, "init_plugin")
        if inspect.ismethod(initializer) and initializer.__self__ is plugin:
            LOG.debug("Initializing plugin class {name}".format(name=plugin))
            initializer()
    return plugin


class PackageDefinition(object):
    def __init__(self, distribution):
        self.name = distribution.project_name
        self.version = distribution.version
        if distribution.has_metadata(distribution.PKG_INFO):
            # This has all the package metadata, including Author,
            # description, License etc
            self.info = distribution.get_metadata(distribution.PKG_INFO)
        else:
            self.info = None
        self.classes = {}


class MuranoPackage(murano_package.MuranoPackage):
    def __init__(self, pkg_loader, package_definition):
        super(MuranoPackage, self).__init__(
            pkg_loader, package_definition.name, runtime_version='1.0')
        for class_name, clazz in six.iteritems(package_definition.classes):
            if hasattr(clazz, "_murano_class_name"):
                LOG.warning(_LW("Class '%(class_name)s' has a MuranoPL "
                                "name '%(name)s' defined which will be "
                                "ignored") %
                            dict(class_name=class_name,
                            name=getattr(clazz, "_murano_class_name")))
            LOG.debug("Registering '%s' from '%s' in class loader",
                      class_name, package_definition.name)
            self.register_class(clazz, class_name)

    def get_resource(self, name):
        raise NotImplementedError()



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from murano.dsl import helpers


class Object(object):
    def __init__(self, __name, __id=None, **kwargs):
        self.data = {
            '?': {
                'type': __name,
                'id': __id or helpers.generate_id()
            }
        }
        self.data.update(kwargs)

    @property
    def id(self):
        return self.data['?']['id']

    @property
    def type_name(self):
        return self.data['?']['type']

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data

    def __delitem__(self, key):
        del self.data[key]


class Attribute(object):
    def __init__(self, obj, key, value):
        self._value = value
        self._key = key
        self._obj = obj

    @property
    def obj(self):
        return self._obj

    @property
    def key(self):
        return self._key

    @property
    def value(self):
        return self._value


class Ref(object):
    def __init__(self, obj):
        if isinstance(obj, six.string_types):
            self._id = obj
        else:
            self._id = obj.id

    @property
    def id(self):
        return self._id


def build_model(root):
    if isinstance(root, dict):
        for key, value in root.items():
            root[key] = build_model(value)
    elif isinstance(root, list):
        for i in range(len(root)):
            root[i] = build_model(root[i])
    elif isinstance(root, Object):
        return build_model(root.data)
    elif isinstance(root, Ref):
        return root.id
    elif isinstance(root, Attribute):
        return [root.obj.id, root.obj.type_name, root.key, root.value]
    return root



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from tempest import config
from tempest.lib import exceptions

from murano_tempest_tests.tests.api.application_catalog import base
from murano_tempest_tests import utils

CONF = config.CONF


class TestStaticActionsNegative(base.BaseApplicationCatalogTest):

    @classmethod
    def resource_setup(cls):
        super(TestStaticActionsNegative, cls).resource_setup()

        application_name = utils.generate_name('test_repository_class')
        cls.abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name, add_class_name=True)

        if CONF.application_catalog.glare_backend:
            client = cls.artifacts_client
        else:
            client = cls.application_catalog_client

        cls.package = client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': False})

    @classmethod
    def resource_cleanup(cls):
        os.remove(cls.abs_archive_path)
        if CONF.application_catalog.glare_backend:
            client = cls.artifacts_client
        else:
            client = cls.application_catalog_client
        client.delete_package(cls.package['id'])
        super(TestStaticActionsNegative, cls).resource_cleanup()

    @testtools.testcase.attr('negative')
    def test_call_static_action_no_args(self):
        self.assertRaises(exceptions.BadRequest,
                          self.application_catalog_client.call_static_action)

    @testtools.testcase.attr('negative')
    def test_call_static_action_wrong_class(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.call_static_action,
                          'wrong.class', 'staticAction',
                          args={'myName': 'John'})

    @testtools.testcase.attr('negative')
    def test_call_static_action_wrong_method(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.call_static_action,
                          class_name=self.package['class_definitions'][0],
                          method_name='wrongMethod',
                          args={'myName': 'John'})

    @testtools.testcase.attr('negative')
    def test_call_static_action_session_method(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.call_static_action,
                          class_name=self.package['class_definitions'][0],
                          method_name='staticNotAction',
                          args={'myName': 'John'})

    @testtools.testcase.attr('negative')
    def test_call_static_action_wrong_args(self):
        self.assertRaises(exceptions.BadRequest,
                          self.application_catalog_client.call_static_action,
                          class_name=self.package['class_definitions'][0],
                          method_name='staticAction',
                          args={'myEmail': 'John'})

    @testtools.testcase.attr('negative')
    def test_call_static_action_wrong_package(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.call_static_action,
                          class_name=self.package['class_definitions'][0],
                          method_name='staticAction',
                          package_name='wrong.package',
                          args={'myName': 'John'})

    @testtools.testcase.attr('negative')
    def test_call_static_action_wrong_version_format(self):
        self.assertRaises(exceptions.BadRequest,
                          self.application_catalog_client.call_static_action,
                          class_name=self.package['class_definitions'][0],
                          method_name='staticAction',
                          class_version='aaa',
                          args={'myName': 'John'})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.common import rpc
from murano.db import models
from murano.db.services import actions as actions_db
from murano.services import states


class ActionServices(object):
    @staticmethod
    def create_action_task(action_name, target_obj,
                           args, environment, session, token):
        action = None
        if action_name and target_obj:
            action = {
                'object_id': target_obj,
                'method': action_name,
                'args': args or {}
            }
        task = {
            'action': action,
            'model': session.description,
            'token': token,
            'tenant_id': environment.tenant_id,
            'id': environment.id
        }
        if session.description['Objects'] is not None:
            task['model']['Objects']['?']['id'] = environment.id
            task['model']['Objects']['applications'] = \
                task['model']['Objects'].pop('services', [])

        return task

    @staticmethod
    def update_task(action, session, task, unit):
        session.state = states.SessionState.deploying
        task_info = models.Task()
        task_info.environment_id = session.environment_id
        task_info.description = dict(session.description.get('Objects'))
        task_info.action = task['action']
        status = models.Status()
        status.text = 'Action {0} is scheduled'.format(action[1]['name'])
        status.level = 'info'
        task_info.statuses.append(status)
        with unit.begin():
            unit.add(session)
            unit.add(task_info)

    @staticmethod
    def submit_task(action_name, target_obj,
                    args, environment, session, token, unit):
        task = ActionServices.create_action_task(
            action_name, target_obj, args,
            environment, session, token)
        task_id = actions_db.update_task(action_name, session, task, unit)
        rpc.engine().handle_task(task)
        return task_id

    @staticmethod
    def execute(action_id, session, unit, token, args=None):
        if args is None:
            args = {}
        environment = actions_db.get_environment(session, unit)
        action = ActionServices.find_action(session.description, action_id)
        if action is None:
            raise LookupError('Action is not found')
        if not action[1].get('enabled', True):
            raise ValueError('Cannot execute disabled action')

        return ActionServices.submit_task(
            action[1]['name'], action[0], args, environment,
            session, token, unit)

    @staticmethod
    def find_action(model, action_id):
        """Find_action for object def with specified action

        Traverses object model looking for an object definition
        containing specified action

        :param model: object model
        :param action_id: ID of an action
        :return: tuple (object id, {"name": "action_name_in_MuranoPL",
                                    "enabled": True })
        """
        if isinstance(model, list):
            for item in model:
                result = ActionServices.find_action(item, action_id)
                if result is not None:
                    return result
        elif isinstance(model, dict):
            if '?' in model and 'id' in model['?'] and \
                    '_actions' in model['?'] and \
                    action_id in model['?']['_actions']:
                return model['?']['id'], model['?']['_actions'][action_id]

            for obj in model.values():
                result = ActionServices.find_action(obj, action_id)
                if result is not None:
                    return result
        else:
            return None

    @staticmethod
    def get_result(environment_id, task_id, unit):
        task = unit.query(models.Task).filter_by(
            id=task_id, environment_id=environment_id).first()

        if task is not None:
            return task.result

        return None



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import identity
from keystoneauth1 import loading as ka_loading
from keystoneclient.v3 import client as ks_client
from oslo_config import cfg

from murano.dsl import helpers


CFG_KEYSTONE_GROUP = 'keystone_authtoken'

cfg.CONF.import_group(CFG_KEYSTONE_GROUP, 'keystonemiddleware.auth_token')


def _get_keystone_auth(trust_id=None):
    if not cfg.CONF[CFG_KEYSTONE_GROUP].auth_type:
        # Fallback to legacy v2 options if no auth_type is set.
        # If auth_type is set, it is possible to use the auth loader
        # from keystoneauth1. This is the same fallback as keystonemiddleware
        # uses.
        kwargs = {
            'auth_url':
                cfg.CONF[CFG_KEYSTONE_GROUP].auth_uri.replace('v2.0', 'v3'),
            'username': cfg.CONF[CFG_KEYSTONE_GROUP].admin_user,
            'password': cfg.CONF[CFG_KEYSTONE_GROUP].admin_password,
            'user_domain_name': "Default"
        }
        if not trust_id:
            kwargs['project_name'] = \
                cfg.CONF[CFG_KEYSTONE_GROUP].admin_tenant_name
            kwargs['project_domain_name'] = "Default"
        else:
            kwargs['trust_id'] = trust_id
        auth = identity.Password(**kwargs)
    else:
        kwargs = {}
        if trust_id:
            # Remove project_name and project_id, since we need a trust scoped
            # auth object
            kwargs['project_name'] = None
            kwargs['project_domain_name'] = None
            kwargs['project_id'] = None
            kwargs['trust_id'] = trust_id
        auth = ka_loading.load_auth_from_conf_options(
            cfg.CONF,
            CFG_KEYSTONE_GROUP,
            **kwargs)
    return auth


def _create_keystone_admin_client():
    auth = _get_keystone_auth()
    session = _get_session(
        auth=auth,
        conf_section=cfg.CONF[CFG_KEYSTONE_GROUP])
    return ks_client.Client(session=session)


def get_client_session(execution_session=None, conf=None):
    if not execution_session:
        execution_session = helpers.get_execution_session()
    trust_id = execution_session.trust_id
    if trust_id is None:
        return get_token_client_session(
            token=execution_session.token,
            project_id=execution_session.project_id)
    auth = _get_keystone_auth(trust_id)
    session = _get_session(auth=auth, conf_section=conf)
    return session


def get_token_client_session(token=None, project_id=None, conf=None):
    auth_url = cfg.CONF[CFG_KEYSTONE_GROUP].auth_uri.replace('v2.0', 'v3')
    if token is None or project_id is None:
        execution_session = helpers.get_execution_session()
        token = execution_session.token
        project_id = execution_session.project_id
    token_auth = identity.Token(
        auth_url,
        token=token,
        project_id=project_id)
    session = _get_session(auth=token_auth, conf_section=conf)
    return session


def create_keystone_client(token=None, project_id=None, conf=None):
    return ks_client.Client(session=get_token_client_session(
        token=token, project_id=project_id, conf=conf))


def create_trust(trustee_token=None, trustee_project_id=None):
    admin_client = _create_keystone_admin_client()
    user_client = create_keystone_client(
        token=trustee_token, project_id=trustee_project_id)
    trustee_user = admin_client.session.auth.get_user_id(admin_client.session)
    auth_ref = user_client.session.auth.get_access(user_client.session)
    trustor_user = auth_ref.user_id
    project = auth_ref.project_id
    roles = auth_ref.role_names
    trust = user_client.trusts.create(
        trustor_user=trustor_user,
        trustee_user=trustee_user,
        impersonation=True,
        role_names=roles,
        project=project)
    return trust.id


def delete_trust(trust):
    user_client = _create_keystone_admin_client()
    user_client.trusts.delete(trust)


def _get_config_option(conf_section, option_names, default=None):
    if not isinstance(option_names, (list, tuple)):
        option_names = (option_names,)
    for name in option_names:
        if hasattr(conf_section, name):
            return getattr(conf_section, name)
    return default


def _get_session(auth, conf_section):
    # Fallback to keystone_authtoken section for TLS parameters
    # if no other conf_section supplied
    if not conf_section:
        conf_section = cfg.CONF[CFG_KEYSTONE_GROUP]
    session = ka_loading.session.Session().load_from_options(
        auth=auth,
        insecure=_get_config_option(conf_section, 'insecure', False),
        cacert=_get_config_option(
            conf_section,
            ('ca_file', 'cafile', 'cacert')),
        key=_get_config_option(conf_section, ('key_file', 'keyfile')),
        cert=_get_config_option(conf_section, ('cert_file', 'certfile')))
    return session


def get_session_client_parameters(service_type=None,
                                  region='',
                                  interface=None,
                                  service_name=None,
                                  conf=None,
                                  session=None,
                                  execution_session=None):
    if region == '':
        region = cfg.CONF.home_region
    result = {
        'session': session or get_client_session(
            execution_session=execution_session, conf=conf)
    }

    url = _get_config_option(conf, 'url')
    if url:
        result['endpoint_override'] = url
    else:
        if not interface:
            interface = _get_config_option(conf, 'endpoint_type')
        result.update({
            'service_type': service_type,
            'service_name': service_name,
            'interface': interface,
            'region_name': region
        })
    return result



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os

import murano.packages.hot_package
import murano.packages.load_utils as load_utils
import murano.tests.unit.base as test_base


class TestHotPackage(test_base.MuranoTestCase):
    def test_heat_files_generated(self):
        package_dir = os.path.abspath(
            os.path.join(__file__,
                         '../../test_packages/test.hot.v1.app_with_files')
        )
        load_utils.load_from_dir(package_dir)

        files = murano.packages.hot_package.HotPackage._translate_files(
            package_dir)
        expected_result = {
            "testHeatFile",
            "middle_file/testHeatFile",
            "middle_file/inner_file/testHeatFile",
            "middle_file/inner_file2/testHeatFile"
        }
        msg = "hot files were not generated correctly"
        self.assertSetEqual(expected_result, set(files), msg)

    def test_heat_files_generated_empty(self):
        package_dir = os.path.abspath(
            os.path.join(__file__,
                         '../../test_packages/test.hot.v1.app')
        )
        load_utils.load_from_dir(package_dir)

        files = murano.packages.hot_package.HotPackage \
            ._translate_files(package_dir)
        msg = "heat files were not generated correctly. Expected empty list"
        self.assertEqual([], files, msg)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import weakref

import six

from murano.dsl import dsl_types
from murano.dsl import helpers
from murano.dsl import murano_object


class ObjectStore(object):
    def __init__(self, executor, parent_store=None):
        self._parent_store = parent_store
        self._store = {}
        self._designer_attributes_store = {}
        self._executor = weakref.ref(executor)

    @property
    def executor(self):
        return self._executor()

    def get(self, object_id):
        result = self._store.get(object_id)
        if not result and self._parent_store:
            return self._parent_store.get(object_id)
        return result

    def has(self, object_id, check_parent_store=True):
        if object_id in self._store:
            return True
        if check_parent_store and self._parent_store:
            return self._parent_store.has(object_id, check_parent_store)
        return False

    def put(self, murano_object, object_id=None):
        self._store[object_id or murano_object.object_id] = murano_object

    def iterate(self):
        return six.iterkeys(self._store)

    def remove(self, object_id):
        self._store.pop(object_id)

    def load(self, value, owner, default_type=None,
             scope_type=None, context=None, keep_ids=False):
        # do the object model load in a temporary object store and copy
        # loaded objects here after that
        model_store = InitializationObjectStore(
            owner, self, keep_ids)
        with helpers.with_object_store(model_store):
            result = model_store.load(
                value, owner, scope_type=scope_type,
                default_type=default_type, context=context)
            for obj_id in model_store.iterate():
                obj = model_store.get(obj_id)
                if obj.initialized:
                    self.put(obj)
            return result

    @staticmethod
    def _get_designer_attributes(header):
        return dict((k, v) for k, v in six.iteritems(header)
                    if str(k).startswith('_'))

    def designer_attributes(self, object_id):
        return self._designer_attributes_store.get(object_id, {})

    @property
    def initializing(self):
        return False

    @property
    def parent_store(self):
        return self._parent_store


# Temporary ObjectStore to load object graphs. Does 2-phase load
# and maintains internal state on what phase is currently running
# as well as objects that are in the middle of initialization.
# Required in order to isolate semi-initialized objects from regular
# objects in main ObjectStore and internal state between graph loads
# in different threads. Once the load is done all objects are copied
# to the parent ObjectStore
class InitializationObjectStore(ObjectStore):
    def __init__(self, root_owner, parent_store, keep_ids):
        super(InitializationObjectStore, self).__init__(
            parent_store.executor, parent_store)
        self._initializing = False
        self._root_owner = root_owner
        self._keep_ids = keep_ids
        self._initializers = []

    @property
    def initializing(self):
        return self._initializing

    def load(self, value, owner, default_type=None,
             scope_type=None, context=None, **kwargs):
        parsed = helpers.parse_object_definition(value, scope_type, context)
        if not parsed:
            raise ValueError('Invalid object representation format')

        if owner is self._root_owner:
            self._initializing = True

        class_obj = parsed['type'] or default_type
        if not class_obj:
            raise ValueError(
                'Invalid object representation: '
                'no type information was provided')
        if isinstance(class_obj, dsl_types.MuranoTypeReference):
            class_obj = class_obj.type
        object_id = parsed['id']
        obj = None if object_id is None else self.get(object_id)
        if not obj:
            obj = murano_object.MuranoObject(
                class_obj, helpers.weak_proxy(owner),
                name=parsed['name'],
                object_id=object_id if self._keep_ids else None)
            self.put(obj, object_id or obj.object_id)

            system_value = ObjectStore._get_designer_attributes(
                parsed['extra'])
            self._designer_attributes_store[object_id] = system_value

        if context is None:
            context = self.executor.create_object_context(obj)

        def run_initialize():
            self._initializers.extend(
                obj.initialize(context, parsed['properties']))

        run_initialize()
        if owner is self._root_owner:
            self._initializing = False
            run_initialize()

        if owner is self._root_owner:
            with helpers.with_object_store(self.parent_store):
                for fn in self._initializers:
                    fn()

        return obj



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os.path
import re

from testtools import matchers

from murano.dsl import dsl_exception
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestExceptions(test_case.DslTestCase):
    def setUp(self):
        super(TestExceptions, self).setUp()

        def exception_func():
            exc = LookupError('just random Python exception')
            frameinfo = inspect.getframeinfo(inspect.currentframe())
            exc._position = \
                os.path.basename(frameinfo.filename), frameinfo.lineno + 4
            # line below must be exactly 4 lines after currentframe()
            raise exc

        self.register_function(exception_func, 'raisePythonException')
        self._runner = self.new_runner(om.Object('ExceptionHandling'))

    def test_throw_catch(self):
        self._runner.testThrow(1)
        self.assertEqual(
            ['enter try', u'exception message', 'finally section'],
            self.traces)

    def test_rethrow(self):
        e = self.assertRaises(
            dsl_exception.MuranoPlException,
            self._runner.testThrow, 2)

        self.assertEqual(['anotherExceptionName'], e.names)
        self.assertEqual('exception message 2', e.message)
        self.assertEqual('[anotherExceptionName]: exception message 2', str(e))

        self.assertEqual(
            ['enter try', 'exception message 2',
             'rethrow', 'finally section'],
            self.traces)

    def test_catch_all_catch(self):
        self._runner.testThrow(3)
        self.assertEqual(
            ['enter try', 'catch all',
             'exception message 3', 'finally section'],
            self.traces)

    def test_no_throw(self):
        self._runner.testThrow(4)
        self.assertEqual(
            ['enter try', 'exit try', 'else section', 'finally section'],
            self.traces)

    def test_stack_trace(self):
        self._runner.preserve_exception = True
        e = self.assertRaises(
            dsl_exception.MuranoPlException,
            self._runner.testStackTrace)
        call_stack = e.format()
        self.assertThat(
            call_stack,
            matchers.StartsWith(
                'LookupError: just random Python exception'))

        self.assertIsInstance(e.original_exception, LookupError)

        filename, line = e.original_exception._position
        self.assertThat(
            call_stack,
            matchers.MatchesRegex(
                r'.*^  File \".*ExceptionHandling\.yaml\", '
                r'line \d+:\d+ in method testStackTrace .*'
                r'of type ExceptionHandling$.*'
                r'^  File \".*{0}\", line {1} '
                r'in method exception_func$.*'.format(filename, line),
                re.MULTILINE | re.DOTALL))



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Increase the size of the text columns storing object model

Revision ID: 013
Revises: 012
Create Date: 2016-04-12 15:46:12.251155

"""

# revision identifiers, used by Alembic.
revision = '013'
down_revision = '012'

from alembic import op
import sqlalchemy as sa
import sqlalchemy.dialects.mysql as sa_mysql

MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    engine = op.get_bind()
    if engine.dialect.dialect_description.startswith('mysql'):
        with op.batch_alter_table('session') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa_mysql.LONGTEXT())
        with op.batch_alter_table('environment') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa_mysql.LONGTEXT())
        with op.batch_alter_table('environment-template') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa_mysql.LONGTEXT())


def downgrade():
    engine = op.get_bind()
    if engine.dialect.dialect_description.startswith('mysql'):
        with op.batch_alter_table('session') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa.TEXT())
        with op.batch_alter_table('environment') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa.TEXT())
        with op.batch_alter_table('environment-template') as batch_op:
            batch_op.alter_column('description',
                                  type_=sa.TEXT())



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from tempest import config

from murano_tempest_tests.tests.api.application_catalog import base
from murano_tempest_tests import utils

CONF = config.CONF


class TestRepositorySanity(base.BaseApplicationCatalogTest):

    @classmethod
    def resource_setup(cls):
        if CONF.application_catalog.glare_backend:
            msg = ("Murano using GLARE backend. "
                   "Repository tests will be skipped.")
            raise cls.skipException(msg)
        super(TestRepositorySanity, cls).resource_setup()

    @testtools.testcase.attr('smoke')
    def test_get_list_packages(self):
        package_list = self.application_catalog_client.get_list_packages()
        self.assertIsInstance(package_list, list)

    @testtools.testcase.attr('smoke')
    def test_upload_and_delete_package(self):
        application_name = utils.generate_name('package_test_upload')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': False})
        package_list = self.application_catalog_client.get_list_packages()
        self.assertIn(package['id'], {pkg['id'] for pkg in package_list})
        self.application_catalog_client.delete_package(package['id'])
        package_list = self.application_catalog_client.get_list_packages()
        self.assertNotIn(package['id'], {pkg['id'] for pkg in package_list})


class TestRepository(base.BaseApplicationCatalogIsolatedAdminTest):

    @classmethod
    def resource_setup(cls):
        if CONF.application_catalog.glare_backend:
            msg = ("Murano using GLARE backend. "
                   "Repository tests will be skipped.")
            raise cls.skipException(msg)

        super(TestRepository, cls).resource_setup()

        application_name = utils.generate_name('test_repository_class')
        cls.abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        cls.package = cls.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': False})

    @classmethod
    def resource_cleanup(cls):
        os.remove(cls.abs_archive_path)
        cls.application_catalog_client.delete_package(cls.package['id'])
        super(TestRepository, cls).resource_cleanup()

    @testtools.testcase.attr('smoke')
    def test_get_package(self):
        package = self.application_catalog_client.get_package(
            self.package['id'])
        self.assertEqual(self.package['tags'], package['tags'])

    @testtools.testcase.attr('smoke')
    def test_update_package(self):
        post_body = [
            {
                "op": "add",
                "path": "/tags",
                "value": ["im a test"]
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertIn("im a test", result['tags'])

        post_body = [
            {
                "op": "replace",
                "path": "/tags",
                "value": ["im bad:D"]
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertNotIn("im a test", result['tags'])
        self.assertIn("im bad:D", result['tags'])

        post_body = [
            {
                "op": "remove",
                "path": "/tags",
                "value": ["im bad:D"]
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertNotIn("im bad:D", result['tags'])

        post_body = [
            {
                "op": "replace",
                "path": "/is_public",
                "value": True
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertTrue(result['is_public'])

        post_body = [
            {
                "op": "replace",
                "path": "/enabled",
                "value": True
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertTrue(result['enabled'])

        post_body = [
            {
                "op": "replace",
                "path": "/description",
                "value": "New description"
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertEqual("New description", result['description'])

        post_body = [
            {
                "op": "replace",
                "path": "/name",
                "value": "New name"
            }
        ]

        result = self.application_catalog_client.update_package(
            self.package['id'], post_body)
        self.assertEqual("New name", result['name'])

    @testtools.testcase.attr('smoke')
    def test_download_package(self):
        self.application_catalog_client.download_package(self.package['id'])

    @testtools.testcase.attr('smoke')
    def test_get_ui_definitions(self):
        self.application_catalog_client.get_ui_definition(self.package['id'])

    @testtools.testcase.attr('smoke')
    def test_get_logo(self):
        self.application_catalog_client.get_logo(self.package['id'])



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import os
import uuid

import eventlet.event
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import base64
import six
from yaql import specs

import murano.common.exceptions as exceptions
from murano.common.messaging import message
from murano.dsl import dsl
import murano.engine.system.common as common

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class AgentException(Exception):
    pass


@dsl.name('io.murano.system.Agent')
class Agent(object):
    def __init__(self, host):
        self._enabled = False
        if CONF.engine.disable_murano_agent:
            LOG.debug('Use of murano-agent is disallowed '
                      'by the server configuration')
            return

        self._host = host
        self._enabled = True
        env = host.find_owner('io.murano.Environment')
        self._queue = str('e%s-h%s' % (env.id, host.id)).lower()

    @property
    def enabled(self):
        return self._enabled

    def prepare(self):
        # (sjmc7) - turn this into a no-op if agents are disabled
        if CONF.engine.disable_murano_agent:
            LOG.debug('Use of murano-agent is disallowed '
                      'by the server configuration')
            return

        region = dsl.MuranoObjectInterface.create(self._host().getRegion())
        with common.create_rmq_client(region) as client:
            client.declare(self._queue, enable_ha=True, ttl=86400000)

    def queue_name(self):
        return self._queue

    def _check_enabled(self):
        if CONF.engine.disable_murano_agent:
            raise exceptions.PolicyViolationException(
                'Use of murano-agent is disallowed '
                'by the server configuration')

    def _prepare_message(self, template, msg_id):
        msg = message.Message()
        msg.body = template
        msg.id = msg_id
        return msg

    def _send(self, template, wait_results, timeout):
        """Send a message over the MQ interface."""
        region = self._host().getRegion()
        msg_id = template.get('ID', uuid.uuid4().hex)
        if wait_results:
            event = eventlet.event.Event()
            listener = region['agentListener']
            listener().subscribe(msg_id, event)

        msg = self._prepare_message(template, msg_id)
        with common.create_rmq_client(region) as client:
            client.send(message=msg, key=self._queue)

        if wait_results:
            try:
                with eventlet.Timeout(timeout):
                    result = event.wait()

            except eventlet.Timeout:
                listener().unsubscribe(msg_id)
                raise exceptions.TimeoutException(
                    'The murano-agent did not respond '
                    'within {0} seconds'.format(timeout))

            if not result:
                return None

            if result.get('FormatVersion', '1.0.0').startswith('1.'):
                return self._process_v1_result(result)

            else:
                return self._process_v2_result(result)

        else:
            return None

    @specs.parameter(
        'resources', dsl.MuranoObjectParameter('io.murano.system.Resources'))
    def call(self, template, resources, timeout=None):
        if timeout is None:
            timeout = CONF.engine.agent_timeout
        self._check_enabled()
        plan = self.build_execution_plan(template, resources())
        return self._send(plan, True, timeout)

    @specs.parameter(
        'resources', dsl.MuranoObjectParameter('io.murano.system.Resources'))
    def send(self, template, resources):
        self._check_enabled()
        plan = self.build_execution_plan(template, resources())
        return self._send(plan, False, 0)

    def call_raw(self, plan, timeout=None):
        if timeout is None:
            timeout = CONF.engine.agent_timeout
        self._check_enabled()
        return self._send(plan, True, timeout)

    def send_raw(self, plan):
        self._check_enabled()
        return self._send(plan, False, 0)

    def is_ready(self, timeout=100):
        try:
            self.wait_ready(timeout)
        except exceptions.TimeoutException:
            return False
        else:
            return True

    def wait_ready(self, timeout=100):
        self._check_enabled()
        template = {'Body': 'return', 'FormatVersion': '2.0.0', 'Scripts': {}}
        self.call(template, False, timeout)

    def _process_v1_result(self, result):
        if result['IsException']:
            raise AgentException(dict(self._get_exception_info(
                result.get('Result', [])), source='execution_plan'))
        else:
            results = result.get('Result', [])
            if not result:
                return None
            value = results[-1]
            if value['IsException']:
                raise AgentException(dict(self._get_exception_info(
                    value.get('Result', [])), source='command'))
            else:
                return value.get('Result')

    def _process_v2_result(self, result):
        error_code = result.get('ErrorCode', 0)
        if not error_code:
            return result.get('Body')
        else:
            body = result.get('Body') or {}
            err = {
                'message': body.get('Message'),
                'details': body.get('AdditionalInfo'),
                'errorCode': error_code,
                'time': result.get('Time')
            }
            for attr in ('Message', 'AdditionalInfo'):
                if attr in body:
                    del body[attr]
            err['extra'] = body if body else None
            raise AgentException(err)

    def _get_array_item(self, array, index):
        return array[index] if len(array) > index else None

    def _get_exception_info(self, data):
        data = data or []
        return {
            'type': self._get_array_item(data, 0),
            'message': self._get_array_item(data, 1),
            'command': self._get_array_item(data, 2),
            'details': self._get_array_item(data, 3),
            'timestamp': datetime.datetime.now().isoformat()
        }

    def build_execution_plan(self, template, resources):
        template = copy.deepcopy(template)
        if not isinstance(template, dict):
            raise ValueError('Incorrect execution plan ')
        format_version = template.get('FormatVersion')
        if not format_version or format_version.startswith('1.'):
            return self._build_v1_execution_plan(template, resources)
        else:
            return self._build_v2_execution_plan(template, resources)

    def _build_v1_execution_plan(self, template, resources):
        scripts_folder = 'scripts'
        script_files = template.get('Scripts', [])
        scripts = []
        for script in script_files:
            script_path = os.path.join(scripts_folder, script)
            script_path = base64.encode_as_text(script_path) + "\n"
            scripts.append(resources.string(script_path, binary=True))
        template['Scripts'] = scripts
        return template

    def _build_v2_execution_plan(self, template, resources):
        scripts_folder = 'scripts'
        plan_id = uuid.uuid4().hex
        template['ID'] = plan_id
        if 'Action' not in template:
            template['Action'] = 'Execute'
        if 'Files' not in template:
            template['Files'] = {}

        files = {}
        for file_id, file_descr in template['Files'].items():
            files[file_descr['Name']] = file_id

        for name, script in template.get('Scripts', {}).items():
            if 'EntryPoint' not in script:
                raise ValueError('No entry point in script ' + name)

            if 'Application' == script['Type']:
                if script['EntryPoint'] not in files:
                    script['EntryPoint'] = self._place_file(
                        scripts_folder, script['EntryPoint'],
                        template, resources, files)
                else:
                    script['EntryPoint'] = files[script['EntryPoint']]
            if 'Files' in script:
                for i, file in enumerate(script['Files']):
                    if self._get_name(file) not in files:
                        script['Files'][i] = self._place_file(
                            scripts_folder, file, template, resources, files)
                    else:
                        script['Files'][i] = files[file]
        return template

    def _is_url(self, file):
        file = self._get_url(file)
        parts = six.moves.urllib.parse.urlsplit(file)
        if not parts.scheme or not parts.netloc:
            return False
        else:
            return True

    def _get_url(self, file):
        if isinstance(file, dict):
            return list(file.values())[0]
        else:
            return file

    def _get_name(self, file):
        if isinstance(file, dict):
            name = list(file.keys())[0]
        else:
            name = file

        if self._is_url(name):
            name = name[name.rindex('/') + 1:len(name)]
        elif name.startswith('<') and name.endswith('>'):
            name = name[1: -1]
        return name

    def _get_file_value(self, file):
        if isinstance(file, dict):
            file = list(file.values())[0]
        return file

    def _get_body(self, file, resources, folder):
        use_base64 = self._is_base64(file)
        if use_base64:
            path = os.path.join(folder, file[1: -1])
            body = resources.string(path, binary=True)
            body = base64.encode_as_text(body) + "\n"
        else:
            path = os.path.join(folder, file)
            body = resources.string(path)
        return body

    def _is_base64(self, file):
        return file.startswith('<') and file.endswith('>')

    def _get_body_type(self, file):
        return 'Base64' if self._is_base64(file) else 'Text'

    def _place_file(self, folder, file, template, resources, files):
        file_value = self._get_file_value(file)
        name = self._get_name(file)
        file_id = uuid.uuid4().hex

        if self._is_url(file_value):
            template['Files'][file_id] = self._get_file_des_downloadable(file)
            files[name] = file_id

        else:
            template['Files'][file_id] = self._get_file_description(
                file, resources, folder)
            files[name] = file_id
        return file_id

    def _get_file_des_downloadable(self, file):
        name = self._get_name(file)
        file = self._get_file_value(file)
        return {
            'Name': str(name),
            'URL': file,
            'Type': 'Downloadable'
        }

    def _get_file_description(self, file, resources, folder):
        name = self._get_name(file)
        file_value = self._get_file_value(file)

        body_type = self._get_body_type(file_value)
        body = self._get_body(file_value, resources, folder)
        return {
            'Name': name,
            'BodyType': body_type,
            'Body': body
        }



#    Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Based on designate/policy.py

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy
from webob import exc as exceptions

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_ENFORCER = None


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
    _ENFORCER = None


def set_rules(data, default_rule=None, overwrite=True):
    default_rule = default_rule or cfg.CONF.policy_default_rule
    if not _ENFORCER:
        LOG.debug("Enforcer not present, recreating at rules stage.")
        init()

    if default_rule:
        _ENFORCER.default_rule = default_rule

    LOG.debug("Loading rules {rules}, default: {def_rule}, overwrite: "
              "{overwrite}".format(rules=data, def_rule=default_rule,
                                   overwrite=overwrite))

    if isinstance(data, dict):
        rules = policy.Rules.from_dict(data, default_rule)
    else:
        rules = policy.Rules.load_json(data, default_rule)

    _ENFORCER.set_rules(rules, overwrite=overwrite)


def init(default_rule=None, use_conf=True):
    global _ENFORCER
    if not _ENFORCER:
        LOG.debug("Enforcer is not present, recreating.")
        _ENFORCER = policy.Enforcer(CONF, use_conf=use_conf)
    _ENFORCER.load_rules()


def check(rule, ctxt, target=None, do_raise=True,
          exc=exceptions.HTTPForbidden):
    if target is None:
        target = {}
    creds = ctxt.to_dict()

    try:
        result = _ENFORCER.enforce(rule, target, creds, do_raise, exc)
    except Exception:
        result = False
        raise
    else:
        return result
    finally:
        extra = {'policy': {'rule': rule, 'target': target}}

        if result:
            LOG.debug("Policy check succeeded for rule {rule} on target "
                      "{target}".format(rule=rule, target=repr(target),
                                        extra=extra))
        else:
            LOG.debug("Policy check failed for rule {rule} on target: "
                      "{target}".format(rule=rule, target=repr(target),
                                        extra=extra))


def check_is_admin(context):
    """Check if the given context is associated with an admin role.

       :param context: Murano request context
       :returns: A non-False value if context role is admin.
    """
    return check('context_is_admin', context,
                 context.to_dict(), do_raise=False)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.services import actions
from murano.tests.unit import base


class TestActionFinder(base.MuranoTestCase):
    def setUp(self):
        super(TestActionFinder, self).setUp()

    def test_simple_root_level_search(self):
        model = {
            '?': {
                'id': 'id1',
                '_actions': {
                    'ad_deploy': {
                        'enabled': True,
                        'name': 'deploy'
                    }
                }
            }
        }
        action = actions.ActionServices.find_action(model, 'ad_deploy')
        self.assertEqual('deploy', action[1]['name'])

    def test_recursive_action_search(self):
        model = {
            '?': {
                'id': 'id1',
                '_actions': {'ad_deploy': {'enabled': True, 'name': 'deploy'}}
            },
            'property': {
                '?': {
                    'id': 'id2',
                    '_actions': {
                        'ad_scale': {'enabled': True, 'name': 'scale'}
                    }
                },
            }
        }
        action = actions.ActionServices.find_action(model, 'ad_scale')
        self.assertEqual('scale', action[1]['name'])



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions as yaql_exceptions
from yaql.language import expressions
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes

from murano.dsl import contracts
from murano.dsl import exceptions
from murano.dsl import helpers


class Check(contracts.ContractMethod):
    name = 'check'

    @specs.parameter('predicate', yaqltypes.YaqlExpression())
    @specs.parameter('msg', yaqltypes.String(nullable=True))
    def __init__(self, engine, predicate, msg=None):
        self.engine = engine
        self.predicate = predicate
        self.msg = msg

    def _call_predicate(self, value):
        context = self.root_context.create_child_context()
        context['$'] = value
        return self.predicate(utils.NO_VALUE, context, self.engine)

    def validate(self):
        if isinstance(self.value, contracts.ObjRef) or self._call_predicate(
                self.value):
            return self.value
        else:
            msg = self.msg
            if not msg:
                msg = "Value {0} doesn't match predicate".format(
                    helpers.format_scalar(self.value))
            raise exceptions.ContractViolationException(msg)

    def transform(self):
        return self.validate()

    def generate_schema(self):
        rest = [True]
        while rest:
            if (isinstance(self.predicate, expressions.BinaryOperator) and
                    self.predicate.operator == 'and'):
                rest = self.predicate.args[1]
                self.predicate = self.predicate.args[0]
            else:
                rest = []
            res = extract_pattern(self.predicate, self.engine,
                                  self.root_context)
            if res is not None:
                self.value.update(res)
            self.predicate = rest
        return self.value


def is_dollar(expr):
    """Check $-expressions in YAQL AST"""
    return (isinstance(expr, expressions.GetContextValue) and
            expr.path.value in ('$', '$1'))


def extract_pattern(expr, engine, context):
    """Translation of certain known patterns of check() contract expressions"""
    if isinstance(expr, expressions.BinaryOperator):
        ops = ('>', '<', '>=', '<=')
        if expr.operator in ops:
            op_index = ops.index(expr.operator)
            if is_dollar(expr.args[0]):
                constant = evaluate_constant(expr.args[1], engine, context)
                if constant is None:
                    return None
            elif is_dollar(expr.args[1]):
                constant = evaluate_constant(expr.args[0], engine, context)
                if constant is None:
                    return None
                op_index = -1 - op_index
            else:
                return None
            op = ops[op_index]
            if op == '>':
                return {'minimum': constant, 'exclusiveMinimum': True}
            elif op == '>=':
                return {'minimum': constant, 'exclusiveMinimum': False}
            if op == '<':
                return {'maximum': constant, 'exclusiveMaximum': True}
            elif op == '<=':
                return {'maximum': constant, 'exclusiveMaximum': False}
        elif expr.operator == 'in' and is_dollar(expr.args[0]):
            lst = evaluate_constant(expr.args[1], engine, context)
            if isinstance(lst, tuple):
                return {'enum': list(lst)}

        elif (expr.operator == '.' and is_dollar(expr.args[0]) and
                isinstance(expr.args[1], expressions.Function)):
            func = expr.args[1]
            if func.name == 'matches':
                constant = evaluate_constant(func.args[0], engine, context)
                if constant is not None:
                    return {'pattern': constant}


def evaluate_constant(expr, engine, context):
    """Evaluate yaql expression into constant value if possible"""
    if isinstance(expr, expressions.Constant):
        return expr.value
    context = context.create_child_context()
    trap = utils.create_marker('trap')
    context['$'] = trap

    @specs.parameter('name', yaqltypes.StringConstant())
    @specs.name('#get_context_data')
    def get_context_data(name, context):
        res = context[name]
        if res is trap:
            raise yaql_exceptions.ResolutionError()
        return res

    context.register_function(get_context_data)

    try:
        return expressions.Statement(expr, engine).evaluate(context=context)
    except yaql_exceptions.YaqlException:
        return None



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from yaql.language import exceptions as yaql_exceptions
from yaql.language import specs
from yaql.language import yaqltypes

from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestStatics(test_case.DslTestCase):
    def setUp(self):
        @dsl.name('test.TestStatics')
        class PythonClass(object):
            @staticmethod
            @specs.parameter('arg', yaqltypes.Integer())
            @specs.inject('receiver', yaqltypes.Receiver())
            def static_python_method(arg, receiver):
                if isinstance(receiver, dsl_types.MuranoObjectInterface):
                    return 3 * arg
                return 7 * arg

            @classmethod
            @specs.inject('receiver', yaqltypes.Receiver())
            def classmethod_python_method(cls, arg, receiver):
                if isinstance(receiver, dsl_types.MuranoObjectInterface):
                    return cls.__name__.upper() + str(arg)
                return cls.__name__ + str(arg)

        super(TestStatics, self).setUp()
        self.package_loader.load_class_package(
            'test.TestStatics', None).register_class(PythonClass)
        self._runner = self.new_runner(
            om.Object('test.TestStatics', staticProperty2='INVALID'))

    def test_call_static_method_on_object(self):
        self.assertEqual(123, self._runner.testCallStaticMethodOnObject())

    def test_call_static_method_on_class_name(self):
        self.assertEqual(123, self._runner.testCallStaticMethodOnClassName())

    def test_call_static_method_on_class_name_with_ns(self):
        self.assertEqual(
            678, self._runner.testCallStaticMethodOnClassNameWithNs())

    def test_call_static_method_from_another_method(self):
        self.assertEqual(
            123 * 5, self._runner.testCallStaticMethodFromAnotherMethod())

    def test_static_this(self):
        self.assertIsInstance(
            self._runner.testStaticThis(), dsl_types.MuranoTypeReference)

    def test_no_access_to_instance_properties(self):
        self.assertRaises(
            exceptions.NoPropertyFound,
            self._runner.testNoAccessToInstanceProperties)

    def test_access_static_property_from_instance_method(self):
        self.assertEqual(
            'xxx', self._runner.testAccessStaticPropertyFromInstanceMethod())

    def test_access_static_property_from_static_method(self):
        self.assertEqual(
            'xxx', self._runner.testAccessStaticPropertyFromStaticMethod())

    def test_modify_static_property_using_dollar(self):
        self.assertEqual(
            'qq', self._runner.testModifyStaticPropertyUsingDollar())

    def test_modify_static_property_using_this(self):
        self.assertEqual(
            'qq', self._runner.testModifyStaticPropertyUsingThis())

    def test_modify_static_property_using_class_name(self):
        self.assertEqual(
            'qq', self._runner.testModifyStaticPropertyUsingClassName())

    def test_modify_static_property_using_ns_class_name(self):
        self.assertEqual(
            'qq', self._runner.testModifyStaticPropertyUsingNsClassName())

    def test_modify_static_property_using_type_func(self):
        self.assertEqual(
            'qq', self._runner.testModifyStaticPropertyUsingTypeFunc())

    def test_property_is_static(self):
        self.assertEqual('qq', self._runner.testPropertyIsStatic())

    def test_static_properties_excluded_from_object_model(self):
        self.assertEqual(
            'staticProperty',
            self._runner.testStaticPropertisNotLoaded())

    def test_type_is_singleton(self):
        self.assertTrue(self._runner.testTypeIsSingleton())

    def test_static_property_inheritance(self):
        self.assertEqual(
            'baseStaticProperty' * 3,
            self._runner.testStaticPropertyInheritance())

    def test_static_property_override(self):
        self.assertEqual(
            [
                'conflictingStaticProperty-child',
                'conflictingStaticProperty-child',
                'conflictingStaticProperty-base',
                'conflictingStaticProperty-child',
                'conflictingStaticProperty-base'
            ], self._runner.testStaticPropertyOverride())

    def test_type_info_of_type(self):
        self.assertTrue(self._runner.testTypeinfoOfType())

    def test_call_python_static_method(self):
        self.assertEqual(
            [333] + [777] * 3,
            self._runner.testCallPythonStaticMethod())

    def test_call_python_classmethod(self):
        self.assertEqual(
            ['PYTHONCLASS!'] + ['PythonClass!'] * 3,
            self._runner.testCallPythonClassMethod())

    def test_call_static_method_on_invalid_class(self):
        self.assertRaises(
            yaql_exceptions.NoMatchingMethodException,
            self._runner.testCallStaticMethodOnInvalidClass)

    def test_static_method_callable_from_python(self):
        self.assertEqual(
            'It works!',
            self._runner.on_class('test.TestStatics').testStaticAction())



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

try:
    # integration with congress is optional
    import congressclient.v1.client as congress_client
except ImportError as congress_client_import_error:
    congress_client = None
from oslo_log import log as logging

from murano.common import auth_utils
from murano.common.i18n import _, _LI
from murano.policy import congress_rules
from murano.policy.modify.actions import action_manager as am


LOG = logging.getLogger(__name__)


class ValidationError(Exception):
    """Raised for validation errors."""
    pass


class ModelPolicyEnforcer(object):
    """Policy Enforcer Implementation using Congress client

    Converts murano model to list of congress data rules.

    We ask congress using simulation api of congress rest client
    to resolve "murano_system:predeploy_errors(env_id, obj_id, msg)"
    table along with congress data rules to return validation results.
    """

    def __init__(self, execution_session, action_manager=None):
        self._execution_session = execution_session
        self._action_manager = action_manager or am.ModifyActionManager()
        self._client = None

    def _create_client(self):
        if not congress_client:
            # congress client was not imported
            raise congress_client_import_error
        return congress_client.Client(
            **auth_utils.get_session_client_parameters(
                service_type='policy',
                execution_session=self._execution_session))

    @property
    def client(self):
        if self._client is None:
            self._client = self._create_client()
        return self._client

    def modify(self, obj, package_loader=None):
        """Modifies model using Congress rule engine.

        @type obj: object model
        @param obj: Representation of model starting on
                      environment level (['Objects'])
        @type class_loader: murano.dsl.class_loader.MuranoClassLoader
        @param class_loader: Optional. Used for evaluating parent class types
        @raises ValidationError in case validation was not successful
        """

        model = obj.to_dictionary()

        LOG.debug('Modifying model')
        LOG.debug(model)

        env_id = model['?']['id']

        result = self._execute_simulation(package_loader, env_id, model,
                                          'predeploy_modify(eid, oid, action)')

        raw_actions = result["result"]
        if raw_actions:
            actions = self._parse_simulation_result('predeploy_modify',
                                                    env_id, raw_actions)
            for action in actions:
                self._action_manager.apply_action(obj, action)

    def validate(self, model, package_loader=None):
        """Validate model using Congress rule engine.

        @type model: dict
        @param model: Dictionary representation of model starting on
                      environment level (['Objects'])
        @type package_loader: murano.dsl.package_loader.MuranoPackageLoader
        @param package_loader: Optional. Used for evaluating parent class types
        @raises ValidationError in case validation was not successful
        """

        if model is None:
            return

        env_id = model['?']['id']

        validation_result = self._execute_simulation(
            package_loader, env_id, model,
            'predeploy_errors(eid, oid, msg)')

        if validation_result["result"]:

            messages = self._parse_simulation_result(
                'predeploy_errors', env_id,
                validation_result["result"])

            if messages:
                result_str = "\n  ".join(map(str, messages))
                msg = _("Murano object model validation failed: {0}").format(
                    "\n  " + result_str)
                LOG.error(msg)
                raise ValidationError(msg)
        else:
            LOG.info(_LI('Model valid'))

    def _execute_simulation(self, package_loader, env_id, model, query):
        rules = congress_rules.CongressRulesManager().convert(
            model, package_loader, self._execution_session.project_id)
        rules_str = list(map(str, rules))
        # cleanup of data populated by murano driver
        rules_str.insert(0, 'deleteEnv("{0}")'.format(env_id))
        rules_line = " ".join(rules_str)
        LOG.debug('Congress rules: \n  {rules} '
                  .format(rules='\n  '.join(rules_str)))

        validation_result = self.client.execute_policy_action(
            "murano_system",
            "simulate",
            False,
            False,
            {'query': query,
             'action_policy': 'murano_action',
             'sequence': rules_line})
        return validation_result

    @staticmethod
    def _parse_simulation_result(query, env_id, results):
        """Transforms the list of results

        Transforms a list of strings in format:

        ['predeploy_errors("env_id_1", "obj_id_1", "message1")',
        'predeploy_errors("env_id_2", "obj_id_2", "message2")']

        to a list of strings with message only filtered to provided
        env_id (e.g. 'env_id_1'):

        ['message2']
        """

        messages = []
        regexp = query + '\("([^"]*)",\s*"([^"]*)",\s*"(.*)"\)'
        for result in results:
            match = re.search(regexp, result)
            if match:
                if env_id in match.group(1):
                    messages.append(match.group(3))

        return messages



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import weakref

import semantic_version
import six
from yaql.language import utils

from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import exceptions
from murano.dsl import helpers
from murano.dsl import meta as dslmeta
from murano.dsl import murano_method
from murano.dsl import murano_object
from murano.dsl import murano_property
from murano.dsl import yaql_integration


class MuranoType(dsl_types.MuranoType):
    def __init__(self, ns_resolver, name, package):
        self._namespace_resolver = ns_resolver
        self._name = name
        self._package = weakref.ref(package)

    @property
    def name(self):
        return self._name

    @property
    def package(self):
        return self._package()

    @property
    def namespace_resolver(self):
        return self._namespace_resolver

    @abc.abstractproperty
    def usage(self):
        raise NotImplementedError()

    @property
    def version(self):
        return self.package.version

    def get_reference(self):
        return dsl_types.MuranoTypeReference(self)


class MuranoClass(dsl_types.MuranoClass, MuranoType, dslmeta.MetaProvider):
    _allowed_usages = {dsl_types.ClassUsages.Class}

    def __init__(self, ns_resolver, name, package, parents, meta=None,
                 imports=None):
        super(MuranoClass, self).__init__(ns_resolver, name, package)
        self._methods = {}
        self._properties = {}
        self._config = {}
        self._extension_class = None
        if (self._name == constants.CORE_LIBRARY_OBJECT or
                parents is utils.NO_VALUE):
            self._parents = []
        else:
            self._parents = parents or [
                package.find_class(constants.CORE_LIBRARY_OBJECT)]
        for p in self._parents:
            if p.usage not in self._allowed_usages:
                raise exceptions.InvalidInheritanceError(
                    u'Type {0} cannot have parent with Usage {1}'.format(
                        self.name, p.usage))
        remappings = self._build_parent_remappings()
        self._parents = self._adjusted_parents(remappings)
        self._context = None
        self._exported_context = None
        self._property_values = {}
        self._meta = dslmeta.MetaData(meta, dsl_types.MetaTargets.Type, self)
        self._meta_values = None
        self._imports = list(self._resolve_imports(imports))

    def _adjusted_parents(self, remappings):
        seen = {}

        def altered_clone(class_):
            seen_class = seen.get(class_)
            if seen_class is not None:
                return seen_class

            cls_remapping = remappings.get(class_)

            if cls_remapping is not None:
                return altered_clone(cls_remapping)

            new_parents = [altered_clone(p) for p in class_._parents]
            if all(a is b for a, b in zip(class_._parents, new_parents)):
                return class_
            res = copy.copy(class_)
            res._parents = new_parents
            res._meta_values = None
            res._context = None
            res._exported_context = None
            seen[class_] = res
            return res
        return [altered_clone(p) for p in self._parents]

    def __eq__(self, other):
        if not isinstance(other, MuranoType):
            return False
        return self.name == other.name and self.version == other.version

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.name, self.version))

    @property
    def usage(self):
        return dsl_types.ClassUsages.Class

    @property
    def parents(self):
        return self._parents

    @property
    def methods(self):
        return self._methods

    @property
    def all_method_names(self):
        names = set(self.methods.keys())
        for c in self.ancestors():
            names.update(c.methods.keys())
        return tuple(names)

    @property
    def extension_class(self):
        return self._extension_class

    @extension_class.setter
    def extension_class(self, cls):
        self._extension_class = cls
        ctor = yaql_integration.get_class_factory_definition(cls, self)
        self.add_method('__init__', ctor)

    def add_method(self, name, payload, original_name=None):
        method = murano_method.MuranoMethod(self, name, payload, original_name)
        self._methods[name] = method
        self._context = None
        self._exported_context = None
        return method

    @property
    def properties(self):
        return self._properties

    @property
    def all_property_names(self):
        names = set(self.properties.keys())
        for c in self.ancestors():
            names.update(c.properties.keys())
        return tuple(names)

    def add_property(self, property_typespec):
        if not isinstance(property_typespec, murano_property.MuranoProperty):
            raise TypeError('property_typespec')
        self._properties[property_typespec.name] = property_typespec

    def _find_symbol_chains(self, func):
        queue = collections.deque([(self, ())])
        while queue:
            cls, path = queue.popleft()
            symbol = func(cls)
            segment = (symbol,) if symbol is not None else ()
            leaf = True
            for p in cls.parents:
                leaf = False
                queue.append((p, path + segment))
            if leaf:
                path += segment
                if path:
                    yield path

    def _resolve_imports(self, imports):
        seen = {self.name}
        for imp in helpers.list_value(imports):
            if imp in seen:
                continue
            type = helpers.resolve_type(imp, self)
            if type in seen:
                continue
            seen.add(imp)
            seen.add(type)
            yield type

    def _choose_symbol(self, func):
        chains = sorted(
            self._find_symbol_chains(func),
            key=lambda t: len(t))
        result = []
        for i in range(len(chains)):
            if chains[i][0] in result:
                continue
            add = True
            for j in range(i + 1, len(chains)):
                common = 0
                if not add:
                    break
                for p in range(len(chains[i])):
                    if chains[i][-p - 1] is chains[j][-p - 1]:
                        common += 1
                    else:
                        break
                if common == len(chains[i]):
                    add = False
                    break
            if add:
                result.append(chains[i][0])
        return result

    def find_method(self, name):
        return self._choose_symbol(lambda cls: cls.methods.get(name))

    def find_property(self, name):
        return self._choose_symbol(
            lambda cls: cls.properties.get(name))

    def find_static_property(self, name):
        def prop_func(cls):
            prop = cls.properties.get(name)
            if prop is not None and prop.usage == 'Static':
                return prop

        result = self._choose_symbol(prop_func)
        if len(result) < 1:
            raise exceptions.NoPropertyFound(name)
        elif len(result) > 1:
            raise exceptions.AmbiguousPropertyNameError(name)
        return result[0]

    def find_single_method(self, name):
        result = self.find_method(name)
        if len(result) < 1:
            raise exceptions.NoMethodFound(name)
        elif len(result) > 1:
            raise exceptions.AmbiguousMethodName(name)
        return result[0]

    def find_methods(self, predicate):
        result = list(filter(predicate, self.methods.values()))
        for c in self.ancestors():
            for method in six.itervalues(c.methods):
                if predicate(method) and method not in result:
                    result.append(method)
        return result

    def find_properties(self, predicate):
        result = list(filter(predicate, self.properties.values()))
        for c in self.ancestors():
            for prop in c.properties.values():
                if predicate(prop) and prop not in result:
                    result.append(prop)
        return result

    def _iterate_unique_methods(self):
        for name in self.all_method_names:
            try:
                yield self.find_single_method(name)
            except exceptions.AmbiguousMethodName as e:
                def func(*args, **kwargs):
                    raise e
                yield murano_method.MuranoMethod(
                    self, name, func, ephemeral=True)

    def find_single_property(self, name):
        result = self.find_property(name)
        if len(result) < 1:
            raise exceptions.NoPropertyFound(name)
        elif len(result) > 1:
            raise exceptions.AmbiguousPropertyNameError(name)
        return result[0]

    def invoke(self, name, this, args, kwargs, context=None):
        method = self.find_single_method(name)
        return method.invoke(this, args, kwargs, context)

    def is_compatible(self, obj):
        if isinstance(obj, (murano_object.MuranoObject,
                            dsl.MuranoObjectInterface,
                            dsl_types.MuranoTypeReference)):
            obj = obj.type
        if obj == self:
            return True
        return any(cls == self for cls in obj.ancestors())

    def __repr__(self):
        return 'MuranoClass({0}/{1})'.format(self.name, self.version)

    def _build_parent_remappings(self):
        """Remaps class parents.

        In case of multiple inheritance class may indirectly get several
        versions of the same class. It is reasonable to try to replace them
        with single version to avoid conflicts. We can do that when within
        versions that satisfy our class package requirements.
        But in order to merge several classes that are not our parents but
        grand parents we will need to modify classes that may be used
        somewhere else (with another set of requirements). We cannot do this.
        So instead we build translation table that will tell which ancestor
        class need to be replaced with which so that we minimize number of
        versions used for single class (or technically packages since version
        is a package attribute). For translation table to work there should
        be a method that returns all class virtual ancestors so that everybody
        will see them instead of accessing class parents directly and getting
        declared ancestors.
        """
        result = {}

        aggregation = {
            self.package.name: {(
                self.package,
                semantic_version.Spec('==' + str(self.package.version))
            )}
        }
        for cls, parent in helpers.traverse(
                ((self, parent) for parent in self._parents),
                lambda cp: ((cp[1], anc) for anc in cp[1].parents)):
            if cls.package != parent.package:
                requirement = cls.package.requirements[parent.package.name]
                aggregation.setdefault(parent.package.name, set()).add(
                    (parent.package, requirement))

        package_bindings = {}
        for versions in six.itervalues(aggregation):
            mappings = self._remap_package(versions)
            package_bindings.update(mappings)

        for cls in helpers.traverse(self.parents, lambda c: c.parents):
            if cls.package in package_bindings:
                package2 = package_bindings[cls.package]
                cls2 = package2.classes[cls.name]
                result[cls] = cls2
        return result

    @staticmethod
    def _remap_package(versions):
        result = {}
        reverse_mappings = {}
        versions_list = sorted(versions, key=lambda x: x[0].version)
        i = 0
        while i < len(versions_list):
            package1, requirement1 = versions_list[i]
            dst_package = None
            for j, (package2, _) in enumerate(versions_list):
                if i == j:
                    continue
                if package2.version in requirement1 and (
                        dst_package is None or
                        dst_package.version < package2.version):
                    dst_package = package2
            if dst_package:
                result[package1] = dst_package
                reverse_mappings.setdefault(dst_package, []).append(package1)
                for package in reverse_mappings.get(package1, []):
                    result[package] = dst_package
                del versions_list[i]
            else:
                i += 1
        return result

    def ancestors(self):
        for c in helpers.traverse(self, lambda t: t.parents):
            if c is not self:
                yield c

    @property
    def context(self):
        if not self._context:
            ctx = None
            for imp in reversed(self._imports):
                if ctx is None:
                    ctx = imp.exported_context
                else:
                    ctx = helpers.link_contexts(ctx, imp.exported_context)

            if ctx is None:
                self._context = yaql_integration.create_empty_context()
            else:
                self._context = ctx.create_child_context()

            for m in self._iterate_unique_methods():
                if m.instance_stub:
                    self._context.register_function(
                        m.instance_stub, name=m.instance_stub.name)
                if m.static_stub:
                    self._context.register_function(
                        m.static_stub, name=m.static_stub.name)
        return self._context

    @property
    def exported_context(self):
        if not self._exported_context:
            self._exported_context = yaql_integration.create_empty_context()
            for m in self._iterate_unique_methods():
                if m.usage == dsl_types.MethodUsages.Extension:
                    if m.instance_stub:
                        self._exported_context.register_function(
                            m.instance_stub, name=m.instance_stub.name)
                    if m.static_stub:
                        self._exported_context.register_function(
                            m.static_stub, name=m.static_stub.name)
        return self._exported_context

    def get_property(self, name, context):
        prop = self.find_static_property(name)
        cls = prop.declaring_type
        value = cls._property_values.get(name, prop.default)
        return prop.transform(value, cls, None, context)

    def set_property(self, name, value, context, dry_run=False):
        prop = self.find_static_property(name)
        cls = prop.declaring_type
        value = prop.transform(value, cls, None, context)
        if not dry_run:
            cls._property_values[name] = value

    def get_meta(self, context):
        if self._meta_values is None:
            executor = helpers.get_executor()
            context = executor.create_type_context(
                self, caller_context=context)
            self._meta_values = dslmeta.merge_providers(
                self, lambda cls: cls._meta, context)
        return self._meta_values


class MuranoMetaClass(dsl_types.MuranoMetaClass, MuranoClass):
    _allowed_usages = {dsl_types.ClassUsages.Meta, dsl_types.ClassUsages.Class}

    def __init__(self, ns_resolver, name, package, parents, meta=None,
                 imports=None):
        super(MuranoMetaClass, self).__init__(
            ns_resolver, name, package, parents, meta, imports)
        self._cardinality = dsl_types.MetaCardinality.One
        self._targets = list(dsl_types.MetaCardinality.All)
        self._inherited = False

    @property
    def usage(self):
        return dsl_types.ClassUsages.Meta

    @property
    def cardinality(self):
        return self._cardinality

    @cardinality.setter
    def cardinality(self, value):
        self._cardinality = value

    @property
    def targets(self):
        return self._targets

    @targets.setter
    def targets(self, value):
        self._targets = value

    @property
    def inherited(self):
        return self._inherited

    @inherited.setter
    def inherited(self, value):
        self._inherited = value

    def __repr__(self):
        return 'MuranoMetaClass({0}/{1})'.format(self.name, self.version)


def create(data, package, name, ns_resolver):
    usage = data.get('Usage', dsl_types.ClassUsages.Class)
    if usage == dsl_types.ClassUsages.Class:
        return _create_class(MuranoClass, name, ns_resolver, data, package)
    elif usage == dsl_types.ClassUsages.Meta:
        return _create_meta_class(
            MuranoMetaClass, name, ns_resolver, data, package)
    else:
        raise ValueError(u'Invalid type Usage: "{}"'.format(usage))


def _create_class(cls, name, ns_resolver, data, package, *args, **kwargs):
    parent_class_names = data.get('Extends')
    parent_classes = []
    if parent_class_names:
        if not utils.is_sequence(parent_class_names):
            parent_class_names = [parent_class_names]
        for parent_name in parent_class_names:
            full_name = ns_resolver.resolve_name(str(parent_name))
            parent_classes.append(package.find_class(full_name))

    type_obj = cls(
        ns_resolver, name, package, parent_classes, data.get('Meta'),
        data.get('Import'), *args, **kwargs)

    properties = data.get('Properties') or {}
    for property_name, property_spec in six.iteritems(properties):
        spec = murano_property.MuranoProperty(
            type_obj, property_name, property_spec)
        type_obj.add_property(spec)

    methods = data.get('Methods') or data.get('Workflow') or {}

    method_mappings = {
        'initialize': '.init',
        'destroy': '.destroy'
    }

    for method_name, payload in six.iteritems(methods):
        type_obj.add_method(
            method_mappings.get(method_name, method_name), payload)

    return type_obj


def _create_meta_class(cls, name, ns_resolver, data, package, *args, **kwargs):
    cardinality = data.get('Cardinality', dsl_types.MetaCardinality.One)
    if cardinality not in dsl_types.MetaCardinality.All:
        raise ValueError(u'Invalid MetaClass Cardinality "{}"'.format(
            cardinality))
    applies_to = data.get('Applies', dsl_types.MetaTargets.All)
    if isinstance(applies_to, six.string_types):
        applies_to = [applies_to]
    if isinstance(applies_to, list):
        applies_to = set(applies_to)
    delta = applies_to - dsl_types.MetaTargets.All - {'All'}
    if delta:
        raise ValueError(u'Invalid MetaClass target(s) {}:'.format(
            ', '.join(map(u'"{}"'.format, delta)))
        )
    if 'All' in applies_to:
        applies_to = dsl_types.MetaTargets.All
    inherited = data.get('Inherited', False)
    if not isinstance(inherited, bool):
        raise ValueError('Invalid Inherited value. Must be true or false')

    meta_cls = _create_class(
        cls, name, ns_resolver, data, package, *args, **kwargs)
    meta_cls.targets = list(applies_to)
    meta_cls.cardinality = cardinality
    meta_cls.inherited = inherited
    return meta_cls


def weigh_type_hierarchy(cls):
    """Weighs classes in type hierarchy by their distance from the root

    :param cls: root of hierarchy
    :return: dictionary that has class name as keys and distance from the root
             a values. Root class has always a distance of 0. If the class
             (or different versions of that class) is achievable through
             several paths the shortest distance is used.
    """

    result = {}
    for c, w in helpers.traverse(
            [(cls, 0)], lambda t: six.moves.map(
                lambda p: (p, t[1] + 1), t[0].parents)):
        result.setdefault(c.name, w)
    return result



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import errno
import functools
import os

from oslo_concurrency import lockutils
from oslo_log import log as logging
from webob import exc

from murano.common.i18n import _
from murano.db import models
from murano.db.services import sessions
from murano.db import session as db_session
from murano.services import states

LOG = logging.getLogger(__name__)


def check_env(request, environment_id):
    unit = db_session.get_session()
    environment = unit.query(models.Environment).get(environment_id)
    if environment is None:
        msg = _('Environment with id {env_id}'
                ' not found').format(env_id=environment_id)
        LOG.error(msg)
        raise exc.HTTPNotFound(explanation=msg)

    if hasattr(request, 'context'):
        if environment.tenant_id != request.context.tenant:
            msg = _('User is not authorized to access'
                    ' these tenant resources')
            LOG.error(msg)
            raise exc.HTTPForbidden(explanation=msg)
    return environment


def check_session(request, environment_id, session, session_id):
    """Validate, that a session is ok."""
    if session is None:
        msg = _('Session <SessionId {id}> is not found').format(id=session_id)
        LOG.error(msg)
        raise exc.HTTPNotFound(explanation=msg)

    if session.environment_id != environment_id:
        msg = _('Session <SessionId {session_id}> is not tied '
                'with Environment <EnvId {environment_id}>').format(
                    session_id=session_id,
                    environment_id=environment_id)
        LOG.error(msg)
        raise exc.HTTPBadRequest(explanation=msg)

    check_env(request, environment_id)


def verify_env(func):
    @functools.wraps(func)
    def __inner(self, request, environment_id, *args, **kwargs):
        check_env(request, environment_id)
        return func(self, request, environment_id, *args, **kwargs)
    return __inner


def verify_env_template(func):
    @functools.wraps(func)
    def __inner(self, request, env_template_id, *args, **kwargs):
        unit = db_session.get_session()
        template = unit.query(models.EnvironmentTemplate).get(env_template_id)
        if template is None:
            msg = _('Environment Template with id {id} not found'
                    ).format(id=env_template_id)
            LOG.error(msg)
            raise exc.HTTPNotFound(explanation=msg)

        if hasattr(request, 'context'):
            if template.tenant_id != request.context.tenant:
                msg = _('User is not authorized to access'
                        ' this tenant resources')
                LOG.error(msg)
                raise exc.HTTPForbidden(explanation=msg)

        return func(self, request, env_template_id, *args, **kwargs)
    return __inner


def verify_session(func):
    @functools.wraps(func)
    def __inner(self, request, *args, **kwargs):
        if hasattr(request, 'context') and not request.context.session:
            msg = _('X-Configuration-Session header which indicates'
                    ' to the session is missed')
            LOG.error(msg)
            raise exc.HTTPBadRequest(explanation=msg)

        session_id = request.context.session

        unit = db_session.get_session()
        session = unit.query(models.Session).get(session_id)

        if session is None:
            msg = _('Session <SessionId {0}> is not found').format(session_id)
            LOG.error(msg)
            raise exc.HTTPNotFound(explanation=msg)

        if not sessions.SessionServices.validate(session):
            msg = _('Session <SessionId {0}> '
                    'is invalid: environment has been updated or '
                    'updating right now with other session').format(session_id)
            LOG.error(msg)
            raise exc.HTTPForbidden(explanation=msg)

        if session.state == states.SessionState.DEPLOYING:
            msg = _('Session <SessionId {0}> is already in deployment state'
                    ).format(session_id)
            raise exc.HTTPForbidden(explanation=msg)
        return func(self, request, *args, **kwargs)
    return __inner


def ensure_tree(path):
    """Create a directory (and any ancestor directories required).

    :param path: Directory to create
    :return: bool, whether the directory has actually been created
    """
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            if not os.path.isdir(path):
                raise
            else:
                return False
        elif e.errno == errno.EISDIR:
            return False
        else:
            raise
    else:
        return True


ExclusiveInterProcessLock = lockutils.InterProcessLock
if os.name == 'nt':
    # no shared locks on windows
    SharedInterProcessLock = lockutils.InterProcessLock
else:
    import fcntl

    class SharedInterProcessLock(lockutils.InterProcessLock):
        def trylock(self):
            # LOCK_EX instead of LOCK_EX
            fcntl.lockf(self.lockfile, fcntl.LOCK_SH | fcntl.LOCK_NB)

        def _do_open(self):
            # the file has to be open in read mode, therefore this method has
            # to be overriden
            basedir = os.path.dirname(self.path)
            if basedir:
                made_basedir = ensure_tree(basedir)
                if made_basedir:
                    self.logger.debug(
                        'Created lock base path `%s`', basedir)
            # The code here is mostly copy-pasted from oslo_concurrency and
            # fasteners. The file has to be open with read permissions to be
            # suitable for shared locking
            if self.lockfile is None or self.lockfile.closed:
                try:
                    # ensure the file is there, but do not obtain an extra file
                    # descriptor, as closing it would release fcntl lock
                    fd = os.open(self.path, os.O_CREAT | os.O_EXCL)
                    os.close(fd)
                except OSError:
                    pass
                self.lockfile = open(self.path, 'r')


class ReaderWriterLock(lockutils.ReaderWriterLock):

    @contextlib.contextmanager
    def write_lock(self, blocking=True):
        """Context manager that grants a write lock.

        Will wait until no active readers. Blocks readers after acquiring.
        Raises a ``RuntimeError`` if an active reader attempts to acquire
        a lock.
        """
        timeout = None if blocking else 0.00001
        me = self._current_thread()
        i_am_writer = self.is_writer(check_pending=False)
        if self.is_reader() and not i_am_writer:
            raise RuntimeError("Reader %s to writer privilege"
                               " escalation not allowed" % me)
        if i_am_writer:
            # Already the writer; this allows for basic reentrancy.
            yield self
        else:
            with self._cond:
                self._pending_writers.append(me)
                while True:
                    # No readers, and no active writer, am I next??
                    if len(self._readers) == 0 and self._writer is None:
                        if self._pending_writers[0] == me:
                            self._writer = self._pending_writers.popleft()
                            break

                    # NOTE(kzaitsev): this actually means, that we can wait
                    # more than timeout times, since if we get True value we
                    # would get another spin inside of the while loop
                    # Should we do anything about it?
                    acquired = self._cond.wait(timeout)
                    if not acquired:
                        yield False
                        return
            try:
                yield True
            finally:
                with self._cond:
                    self._writer = None
                    self._cond.notify_all()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for database migrations.
For the opportunistic testing you need to set up a db named 'openstack_citest'
with user 'openstack_citest' and password 'openstack_citest' on localhost.
The test will then use that db and u/p combo to run the tests.
For postgres on Ubuntu this can be done with the following commands:
sudo -u postgres psql
postgres=# create user openstack_citest with createdb login password
      'openstack_citest';
postgres=# create database openstack_citest with owner openstack_citest;
"""

import datetime
import uuid

from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import test_base
from oslo_db.sqlalchemy import utils as db_utils
import sqlalchemy

from murano.db.migration import migration
from murano.tests.unit.db.migration import test_migrations_base as base


class MuranoMigrationsCheckers(object):

    snake_walk = True
    downgrade = True

    def assertColumnExists(self, engine, table, column):
        t = db_utils.get_table(engine, table)
        self.assertIn(column, t.c)

    def assertColumnsExists(self, engine, table, columns):
        for column in columns:
            self.assertColumnExists(engine, table, column)

    def assertColumnCount(self, engine, table, columns):
        t = db_utils.get_table(engine, table)
        self.assertEqual(len(t.columns), len(columns))

    def assertColumnNotExists(self, engine, table, column):
        t = db_utils.get_table(engine, table)
        self.assertNotIn(column, t.c)

    def assertIndexExists(self, engine, table, index):
        t = db_utils.get_table(engine, table)
        index_names = [idx.name for idx in t.indexes]
        self.assertIn(index, index_names)

    def assertColumnType(self, engine, table, column, sqltype):
        t = db_utils.get_table(engine, table)
        col = getattr(t.c, column)
        self.assertIsInstance(col.type, sqltype)

    def assertIndexMembers(self, engine, table, index, members):
        self.assertIndexExists(engine, table, index)

        t = db_utils.get_table(engine, table)
        index_columns = None
        for idx in t.indexes:
            if idx.name == index:
                index_columns = idx.columns.keys()
                break

        self.assertEqual(sorted(members), sorted(index_columns))

    def test_walk_versions(self):
        self.walk_versions(self.engine, self.snake_walk, self.downgrade)

    def _check_001(self, engine, data):
        self.assertEqual('001', migration.version(engine))
        self.assertColumnExists(engine, 'category', 'id')
        self.assertColumnExists(engine, 'environment', 'tenant_id')

        # make sure indexes are in place
        self.assertIndexExists(engine,
                               'class_definition',
                               'ix_class_definition_name')

        self.assertIndexExists(engine,
                               'package',
                               'ix_package_fqn')
        self.assertIndexExists(engine,
                               'category',
                               'ix_category_name')

        self._test_package_fqn_is_uniq(engine)

    def _test_package_fqn_is_uniq(self, engine):
        package_table = db_utils.get_table(engine, 'package')

        package = {
            'id': str(uuid.uuid4()),
            'archive': b"archive blob here",
            'fully_qualified_name': 'com.example.package',
            'type': 'class',
            'author': 'OpenStack',
            'name': 'package',
            'enabled': True,
            'description': 'some text',
            'is_public': False,
            'tags': ['tag1', 'tag2'],
            'logo': b"logo blob here",
            'ui_definition': '{}',
            'owner_id': '123',
            'created': datetime.datetime.now(),
            'updated': datetime.datetime.now()
        }
        package_table.insert().execute(package)

        package['id'] = str(uuid.uuid4())
        self.assertRaises(db_exc.DBDuplicateEntry,
                          package_table.insert().execute, package)

    def _check_002(self, engine, data):
        self.assertEqual('002', migration.version(engine))
        self.assertColumnsExists(engine,
                                 'package',
                                 ['supplier_logo', 'supplier'])

    def _check_003(self, engine, data):
        self.assertEqual('003', migration.version(engine))
        self.assertColumnExists(engine, 'task', 'action')
        self.assertColumnExists(engine, 'status', 'task_id')

    def _check_004(self, engine, data):
        self.assertEqual('004', migration.version(engine))
        self.assertColumnType(engine,
                              'package',
                              'description',
                              sqlalchemy.Text)

    def _check_005(self, engine, data):
        self.assertEqual('005', migration.version(engine))
        self.assertColumnExists(engine, 'environment-template', 'id')
        self.assertColumnExists(engine, 'environment-template', 'tenant_id')
        self.assertColumnExists(engine, 'environment-template', 'name')

    def _check_006(self, engine, data):
        self.assertEqual('006', migration.version(engine))
        self.assertColumnExists(engine, 'task', 'result')

    def _check_007(self, engine, data):
        self.assertEqual('007', migration.version(engine))
        self.assertColumnExists(engine, 'locks', 'id')
        self.assertColumnExists(engine, 'locks', 'ts')

    def _check_008(self, engine, data):
        self.assertEqual('008', migration.version(engine))
        self.assertIndexExists(engine,
                               'class_definition',
                               'ix_class_definition_name')

        self.assertIndexExists(engine,
                               'package',
                               'ix_package_fqn_and_owner')


class TestMigrationsMySQL(MuranoMigrationsCheckers,
                          base.BaseWalkMigrationTestCase,
                          test_base.MySQLOpportunisticTestCase):
    pass


class TestMigrationsPostgresql(MuranoMigrationsCheckers,
                               base.BaseWalkMigrationTestCase,
                               test_base.PostgreSQLOpportunisticTestCase):
    pass



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from tempest.lib import exceptions

from murano_tempest_tests.tests.api.application_catalog import base
from murano_tempest_tests import utils


class TestCategoriesNegative(base.BaseApplicationCatalogIsolatedAdminTest):

    @classmethod
    def resource_setup(cls):
        super(TestCategoriesNegative, cls).resource_setup()
        application_name = utils.generate_name(cls.__name__)
        name = utils.generate_name(cls.__name__)
        cls.category = cls.application_catalog_client.create_category(name)
        cls.abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        cls.package = cls.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [cls.category['name']],
             "tags": [], 'is_public': False})

    @classmethod
    def resource_cleanup(cls):
        os.remove(cls.abs_archive_path)
        cls.application_catalog_client.delete_package(cls.package['id'])
        cls.application_catalog_client.delete_category(cls.category['id'])
        super(TestCategoriesNegative, cls).resource_cleanup()

    @testtools.testcase.attr('negative')
    def test_delete_category_by_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.delete_category,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_get_category_by_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.get_category,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_create_category_with_same_name(self):
        self.assertRaises(exceptions.Conflict,
                          self.application_catalog_client.create_category,
                          self.category['name'])

    @testtools.testcase.attr('negative')
    def test_delete_category_with_package(self):
        self.assertRaises(exceptions.Forbidden,
                          self.application_catalog_client.delete_category,
                          self.category['id'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import alembic
from alembic import config as alembic_config
from alembic import migration as alembic_migration

from murano.db import session as db_session


def get_alembic_config():
    path = os.path.join(os.path.dirname(__file__), 'alembic.ini')

    config = alembic_config.Config(path)
    config.set_main_option('script_location',
                           'murano.db.migration:alembic_migrations')
    return config


def version(engine=None):
    """Returns current database version."""
    engine = engine or db_session.get_engine()
    with engine.connect() as conn:
        context = alembic_migration.MigrationContext.configure(conn)
        return context.get_current_revision()


def upgrade(revision, config=None):
    """Used for upgrading database.

    :param version: Desired database version
    :type version: string
    """
    revision = revision or 'head'
    config = config or get_alembic_config()

    alembic.command.upgrade(config, revision or 'head')


def downgrade(revision, config=None):
    """Used for downgrading database.

    :param version: Desired database version7
    :type version: string
    """
    revision = revision or 'base'
    config = config or get_alembic_config()
    return alembic.command.downgrade(config, revision)


def stamp(revision, config=None):
    """Stamps database with provided revision.

    Don't run any migrations.

    :param revision: Should match one from repository or head - to stamp
                     database with most recent revision
    :type revision: string
    """
    config = config or get_alembic_config()
    return alembic.command.stamp(config, revision=revision)


def revision(message=None, autogenerate=False, config=None):
    """Creates template for migration.

    :param message: Text that will be used for migration title
    :type message: string
    :param autogenerate: If True - generates diff based on current database
                         state
    :type autogenerate: bool
    """
    config = config or get_alembic_config()
    return alembic.command.revision(config, message=message,
                                    autogenerate=autogenerate)



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from murano.dsl import helpers


class CongressRulesManager(object):
    """Converts murano model to list of congress rules

    The Congress rules are:
        - murano:objects+(env_id, obj_id, type_name)
        - murano:properties+(obj_id, prop_name, prop_value)
        - murano:relationships+(source, target, name)
        - murano:parent_types+(obj_id, parent_name)
        - murano:states+(env_id, state)
    """

    _rules = []
    _env_id = ''
    _package_loader = None

    def convert(self, model, package_loader=None, tenant_id=None):
        self._rules = []
        self._package_loader = package_loader

        if model is None:
            return self._rules

        self._env_id = model['?']['id']

        state_rule = StateRule(self._env_id, 'pending')
        self._rules.append(state_rule)

        self._walk(model, owner_id=tenant_id)

        # Convert MuranoProperty containing reference to another object
        # to MuranoRelationship.
        object_ids = [rule.obj_id for rule in self._rules
                      if isinstance(rule, ObjectRule)]

        self._rules = [self._create_relationship(rule, object_ids)
                       for rule in self._rules]

        relations = [(rel.source_id, rel.target_id)
                     for rel in self._rules
                     if isinstance(rel, RelationshipRule)]
        closure = self.transitive_closure(relations)

        for rel in closure:
            self._rules.append(ConnectedRule(rel[0], rel[1]))

        return self._rules

    @staticmethod
    def transitive_closure(relations):
        """Computes transitive closure on a directed graph.

        In other words computes reachability within the graph.
        E.g. {(1, 2), (2, 3)} -> {(1, 2), (2, 3), (1, 3)}
        (1, 3) was added because there is path from 1 to 3 in the graph.

        :param relations: list of relations/edges in form of tuples
        :return: transitive closure including original relations
        """
        closure = set(relations)
        while True:
            # Attempts to discover new transitive relations
            # by joining 2 subsequent relations/edges within the graph.
            new_relations = {(x, w) for x, y in closure
                             for q, w in closure if q == y}
            # Creates union with already discovered relations.
            closure_until_now = closure | new_relations
            # If no new relations were discovered in last cycle
            # the computation is finished.
            if closure_until_now == closure:
                return closure
            closure = closure_until_now

    def _walk(self, obj, owner_id, path=()):

        if obj is None:
            return

        obj = self._to_dict(obj)
        new_owner = self._process_item(obj, owner_id, path) or owner_id
        if isinstance(obj, list) or isinstance(obj, tuple):
            for v in obj:
                self._walk(v, new_owner, path)
        elif isinstance(obj, dict):
            for key, value in six.iteritems(obj):
                self._walk(value, new_owner, path + (key, ))

    def _process_item(self, obj, owner_id, path):
        if isinstance(obj, dict) and '?' in obj:
            obj_rule = self._create_object_rule(obj, owner_id)

            self._rules.append(obj_rule)
            # the environment has 'services' relationships
            # to all its top-level applications
            # traversal path is used to test whether
            # we are at the right place within the tree
            if path == ('applications',):
                self._rules.append(RelationshipRule(self._env_id,
                                                    obj_rule.obj_id,
                                                    "services"))
            self._rules.extend(
                self._create_property_rules(obj_rule.obj_id, obj))

            cls = obj['?']['type']
            if 'classVersion' in obj['?']:
                version_spec = obj['?']['classVersion']
            else:
                version_spec = '*'
            types = self._get_parent_types(
                cls, self._package_loader, version_spec)
            self._rules.extend(self._create_parent_type_rules(obj['?']['id'],
                                                              types))
            # current object will be the owner for its subtree
            return obj_rule.obj_id

    @staticmethod
    def _to_dict(obj):
        # If we have MuranoObject class we need to convert to dictionary.
        if 'to_dictionary' in dir(obj):
            return obj.to_dictionary()
        else:
            return obj

    def _create_object_rule(self, app, owner_id):
        return ObjectRule(app['?']['id'], owner_id, app['?']['type'])

    def _create_property_rules(self, obj_id, obj, prefix=""):
        rules = []

        # Skip when inside properties of other object.
        if '?' in obj and prefix != "":
            rules.append(RelationshipRule(obj_id, obj['?']['id'],
                                          prefix.split('.')[0]))
            return rules

        for key, value in six.iteritems(obj):
            if key == '?':
                continue

            if value is not None:
                value = self._to_dict(value)
                if isinstance(value, dict):
                    rules.extend(self._create_property_rules(
                        obj_id, value, prefix + key + "."))
                elif isinstance(value, list) or isinstance(obj, tuple):
                    for v in value:
                        v = self._to_dict(v)
                        if not isinstance(v, dict):
                            rule = PropertyRule(obj_id, prefix + key, v)
                            rules.append(rule)
                else:
                    rule = PropertyRule(obj_id, prefix + key, value)
                    rules.append(rule)

        return rules

    @staticmethod
    def _is_relationship(rule, app_ids):
        if not isinstance(rule, PropertyRule):
            return False

        return rule.prop_value in app_ids

    def _create_relationship(self, rule, app_ids):
        if self._is_relationship(rule, app_ids):
            return RelationshipRule(rule.obj_id, rule.prop_value,
                                    rule.prop_name)
        else:
            return rule

    @staticmethod
    def _get_parent_types(type_name, package_loader, version_spec):
        type_name, version_spec, _ = helpers.parse_type_string(
            type_name, version_spec, None)
        version_spec = helpers.parse_version_spec(version_spec)
        result = {type_name}
        if package_loader:
            pkg = package_loader.load_class_package(type_name, version_spec)
            cls = pkg.find_class(type_name, False)
            if cls:
                result.update(t.name for t in cls.ancestors())
        return result

    @staticmethod
    def _create_parent_type_rules(app_id, types):
        rules = []
        for type_name in types:
            rules.append(ParentTypeRule(app_id, type_name))
        return rules


class ObjectRule(object):
    def __init__(self, obj_id, owner_id, type_name):
        self.obj_id = obj_id
        self.owner_id = owner_id
        self.type_name = helpers.parse_type_string(type_name, None, None)[0]

    def __str__(self):
        return 'murano:objects+("{0}", "{1}", "{2}")'.format(self.obj_id,
                                                             self.owner_id,
                                                             self.type_name)


class PropertyRule(object):
    def __init__(self, obj_id, prop_name, prop_value):
        self.obj_id = obj_id
        self.prop_name = prop_name
        self.prop_value = prop_value

    def __str__(self):
        return 'murano:properties+("{0}", "{1}", "{2}")'.format(
            self.obj_id, self.prop_name, self.prop_value)


class RelationshipRule(object):
    def __init__(self, source_id, target_id, rel_name):
        self.source_id = source_id
        self.target_id = target_id
        self.rel_name = rel_name

    def __str__(self):
        return 'murano:relationships+("{0}", "{1}", "{2}")'.format(
            self.source_id, self.target_id, self.rel_name)


class ConnectedRule(object):
    def __init__(self, source_id, target_id):
        self.source_id = source_id
        self.target_id = target_id

    def __str__(self):
        return 'murano:connected+("{0}", "{1}")'.format(
            self.source_id, self.target_id)


class ParentTypeRule(object):
    def __init__(self, obj_id, type_name):
        self.obj_id = obj_id
        self.type_name = type_name

    def __str__(self):
        return 'murano:parent_types+("{0}", "{1}")'.format(self.obj_id,
                                                           self.type_name)


class StateRule(object):
    def __init__(self, obj_id, state):
        self.obj_id = obj_id
        self.state = state

    def __str__(self):
        return 'murano:states+("{0}", "{1}")'.format(self.obj_id, self.state)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oslo_config import cfg
import oslo_messaging as messaging
from oslo_utils import timeutils
import six

from murano.common import uuidutils
from murano.dsl import dsl

CONF = cfg.CONF


@dsl.name('io.murano.system.StatusReporter')
class StatusReporter(object):
    transport = None

    def __init__(self, environment):
        if StatusReporter.transport is None:
            StatusReporter.transport = messaging.get_transport(CONF)
        self._notifier = messaging.Notifier(
            StatusReporter.transport,
            publisher_id=uuidutils.generate_uuid(),
            topic='murano')
        if isinstance(environment, six.string_types):
            self._environment_id = environment
        else:
            self._environment_id = environment.id

    def _report(self, instance, msg, details=None, level='info'):
        body = {
            'id': (self._environment_id if instance is None
                   else instance.id),
            'text': msg,
            'details': details,
            'level': level,
            'environment_id': self._environment_id,
            'timestamp': timeutils.isotime(subsecond=True)
        }
        self._notifier.info({}, 'murano.report_notification', body)

    def report(self, instance, msg):
        self._report(instance, msg)

    def report_error_(self, instance, msg):
        self._report(instance, msg, None, 'error')

    @dsl.name('report_error')
    def report_error(self, instance, msg):
        self._report(instance, msg, None, 'error')



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_messaging import target
from oslo_service import service
from oslo_utils import timeutils
import pytz
from sqlalchemy import desc

from murano.common.helpers import token_sanitizer
from murano.db import models
from murano.db.services import environments
from murano.db.services import instances
from murano.db import session
from murano.common.i18n import _LI, _LW
from murano.services import states

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class ResultEndpoint(object):
    @staticmethod
    def process_result(context, result, environment_id):
        secure_result = token_sanitizer.TokenSanitizer().sanitize(result)
        LOG.debug('Got result from orchestration '
                  'engine:\n{result}'.format(result=secure_result))

        model = result['model']
        action_result = result.get('action', {})

        unit = session.get_session()
        environment = unit.query(models.Environment).get(environment_id)

        if not environment:
            LOG.warning(_LW('Environment result could not be handled, '
                            'specified environment not found in database'))
            return

        if model['Objects'] is None and model.get('ObjectsCopy', {}) is None:
            environments.EnvironmentServices.remove(environment_id)
            return

        environment.description = model
        if environment.description['Objects'] is not None:
            environment.description['Objects']['services'] = \
                environment.description['Objects'].pop('applications', [])
            action_name = 'Deployment'
            deleted = False
        else:
            action_name = 'Deletion'
            deleted = True
        environment.version += 1
        environment.save(unit)

        # close deployment
        deployment = get_last_deployment(unit, environment.id)
        deployment.finished = timeutils.utcnow()
        deployment.result = action_result

        num_errors = unit.query(models.Status)\
            .filter_by(level='error', task_id=deployment.id).count()
        num_warnings = unit.query(models.Status)\
            .filter_by(level='warning', task_id=deployment.id).count()

        final_status_text = action_name + ' finished'
        if num_errors:
            final_status_text += " with errors"

        elif num_warnings:
            final_status_text += " with warnings"

        status = models.Status()
        status.task_id = deployment.id
        status.text = final_status_text
        status.level = 'info'
        deployment.statuses.append(status)
        deployment.save(unit)

        # close session
        conf_session = unit.query(models.Session).filter_by(
            **{'environment_id': environment.id,
               'state': states.SessionState.DEPLOYING if not deleted
               else states.SessionState.DELETING}).first()
        if num_errors > 0 or result['action'].get('isException'):
            conf_session.state = \
                states.SessionState.DELETE_FAILURE if deleted else \
                states.SessionState.DEPLOY_FAILURE
        else:
            conf_session.state = states.SessionState.DEPLOYED
        conf_session.save(unit)

        # output application tracking information
        services = []
        objects = model['Objects']
        if objects:
            services = objects.get('services')
        if num_errors + num_warnings > 0:
            LOG.warning(_LW('EnvId: {env_id} TenantId: {tenant_id} Status: '
                            'Failed Apps: {services}')
                        .format(env_id=environment.id,
                                tenant_id=environment.tenant_id,
                                services=services))
        else:
            LOG.info(_LI('EnvId: {env_id} TenantId: {tenant_id} Status: '
                         'Successful Apps: {services}')
                     .format(env_id=environment.id,
                             tenant_id=environment.tenant_id,
                             services=services))


def notification_endpoint_wrapper(priority='info'):
    def wrapper(func):
        class NotificationEndpoint(object):
            def __init__(self):
                setattr(self, priority, self._handler)

            def _handler(self, ctxt, publisher_id, event_type,
                         payload, metadata):
                if event_type == ('murano.%s' % func.__name__):
                    func(payload)

            def __call__(self, payload):
                return func(payload)
        return NotificationEndpoint()
    return wrapper


@notification_endpoint_wrapper()
def track_instance(payload):
    LOG.debug('Got track instance request from orchestration '
              'engine:\n{payload}'.format(payload=payload))
    instance_id = payload['instance']
    instance_type = payload.get('instance_type', 0)
    environment_id = payload['environment']
    unit_count = payload.get('unit_count')
    type_name = payload['type_name']
    type_title = payload.get('type_title')

    instances.InstanceStatsServices.track_instance(
        instance_id, environment_id, instance_type,
        type_name, type_title, unit_count)


@notification_endpoint_wrapper()
def untrack_instance(payload):
    LOG.debug('Got untrack instance request from orchestration '
              'engine:\n{payload}'.format(payload=payload))
    instance_id = payload['instance']
    environment_id = payload['environment']
    instances.InstanceStatsServices.destroy_instance(
        instance_id, environment_id)


@notification_endpoint_wrapper()
def report_notification(report):
    LOG.debug('Got report from orchestration '
              'engine:\n{report}'.format(report=report))

    report['entity_id'] = report.pop('id')

    status = models.Status()
    if 'timestamp' in report:
        dt = timeutils.parse_isotime(report.pop('timestamp'))
        report['created'] = dt.astimezone(pytz.utc).replace(tzinfo=None)
    status.update(report)

    unit = session.get_session()
    # connect with deployment
    with unit.begin():
        running_deployment = get_last_deployment(unit,
                                                 status.environment_id)
        status.task_id = running_deployment.id
        unit.add(status)


def get_last_deployment(unit, env_id):
    query = unit.query(models.Task) \
        .filter_by(environment_id=env_id) \
        .order_by(desc(models.Task.started))
    return query.first()


class Service(service.Service):
    """Service class, that contains common methods for custom services"""

    def __init__(self):
        super(Service, self).__init__()
        self.server = None

    def stop(self, graceful=False):
        if self.server:
            self.server.stop()
            if graceful:
                self.server.wait()
        super(Service, self).stop()

    def reset(self):
        if self.server:
            self.server.reset()
        super(Service, self).reset()


class NotificationService(Service):
    def __init__(self):
        super(NotificationService, self).__init__()
        self.server = None

    def start(self):
        endpoints = [report_notification, track_instance, untrack_instance]

        transport = messaging.get_transport(CONF)
        s_target = target.Target(topic='murano', server=str(uuid.uuid4()))

        self.server = messaging.get_notification_listener(
            transport, [s_target], endpoints, executor='eventlet')

        self.server.start()
        super(NotificationService, self).start()


class ApiService(Service):

    def start(self):
        endpoints = [ResultEndpoint()]

        transport = messaging.get_transport(CONF)
        s_target = target.Target('murano', 'results', server=str(uuid.uuid4()))
        self.server = messaging.get_rpc_server(
            transport, s_target, endpoints, 'eventlet')
        self.server.start()
        super(ApiService, self).start()



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging

from murano.common.i18n import _LE

LOG = logging.getLogger(__name__)


class ExecutionSession(object):
    def __init__(self):
        self.token = None
        self.project_id = None
        self.trust_id = None
        self.system_attributes = {}
        self._set_up_list = []
        self._tear_down_list = []

    def on_session_start(self, delegate):
        self._set_up_list.append(delegate)

    def on_session_finish(self, delegate):
        self._tear_down_list.append(delegate)

    def start(self):
        for delegate in self._set_up_list:
            try:
                delegate()
            except Exception:
                LOG.exception(_LE('Unhandled exception on invocation of '
                              'pre-execution hook'))
        self._set_up_list = []

    def finish(self):
        for delegate in self._tear_down_list:
            try:
                delegate()
            except Exception:
                LOG.exception(_LE('Unhandled exception on invocation of '
                              'post-execution hook'))
        self._tear_down_list = []



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import six
import testtools

from tempest import config

from murano_tempest_tests.tests.api.application_catalog.artifacts import base
from murano_tempest_tests import utils

CONF = config.CONF


class TestVersioningNegative(base.BaseArtifactsTest):

    @classmethod
    def resource_setup(cls):
        if not CONF.application_catalog.glare_backend:
            msg = ("Murano is not using GLARE backend. "
                   "Skipping GLARE tests.")
            raise cls.skipException(msg)
        super(TestVersioningNegative, cls).resource_setup()

        # create package with version 1.0.0
        application_name = utils.generate_name('package_test')
        provided_version = '1.0.0'
        package1, path1 = cls.upload_package(
            application_name, version=provided_version)

        # main application
        expected_version = '2.0.0'
        main_app_name = utils.generate_name('main_package_test')
        require = [(package1['name'], expected_version)]
        package2, path2 = cls.upload_package(main_app_name, require=require)

        cls.packages = {
            '1.0.0': package1,
            'require_for_1.0.0': package2
        }
        cls.abs_archive_paths = [path1, path2]

    @classmethod
    def resource_cleanup(cls):
        for pkg in six.itervalues(cls.packages):
            cls.artifacts_client.delete_package(pkg['id'])
        map(os.remove, cls.abs_archive_paths)
        super(TestVersioningNegative, cls).resource_cleanup()

    @testtools.testcase.attr('negative')
    @testtools.testcase.attr('smoke')
    def test_deploy_package_with_no_required_package_version(self):
        """Test deployment of package which requires package with absent version.

        1) Create environment.
        2) Add to the environment package which requires version 2.0.0 of the
        package, which is present with version 1.0.0 only in repository.
        3) Deploy environment.
        4) Check if deployment status failure.
        """

        # create environment
        environment_name = utils.generate_name('create_environment')
        environment = self.application_catalog_client.create_environment(
            environment_name)
        self.addCleanup(self.application_catalog_client.delete_environment,
                        environment['id'])

        # create session
        session = self.application_catalog_client.create_session(
            environment['id'])

        object_model = self.create_obj_model(
            self.packages['require_for_1.0.0'])

        self.application_catalog_client.create_service(
            environment['id'], session['id'], object_model)

        self.application_catalog_client.deploy_session(
            environment['id'], session['id'])

        deploy_result = utils.wait_for_environment_deploy(
            self.application_catalog_client, environment['id'])['status']

        self.assertEqual(deploy_result, 'deploy failure')



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from murano.dsl import constants
from murano.dsl import dsl_exception
from murano.dsl import expressions
from murano.dsl import helpers
from murano.dsl import macros
from murano.dsl.principal_objects import stack_trace
from murano.dsl import yaql_integration


class ThrowMacro(expressions.DslExpression):
    def __init__(self, Throw, Message=None, Extra=None, Cause=None):
        if not Throw:
            raise ValueError()
        if not isinstance(Throw, list):
            Throw = [Throw]

        self._names = Throw
        self._message = Message
        self._extra = Extra or {}
        self._cause = Cause

    def _resolve_names(self, names, context):
        murano_class = helpers.get_type(context)
        for name in names:
            yield murano_class.namespace_resolver.resolve_name(name)

    def execute(self, context):
        stacktrace = stack_trace.create_stack_trace(context, False)
        cause = None
        if self._cause:
            cause = helpers.evaluate(self._cause, context).get_property(
                'nativeException')
        raise dsl_exception.MuranoPlException(
            list(self._resolve_names(helpers.evaluate(self._names, context),
                                     context)),
            helpers.evaluate(self._message, context),
            stacktrace, self._extra, cause)

    def __unicode__(self):
        if self._message:
            return u'Throw {0}: {1}'.format(self._names, self._message)
        return u'Throw ' + six.text_type(self._names)


class CatchBlock(expressions.DslExpression):
    def __init__(self, With=None, As=None, Do=None):
        if With is not None and not isinstance(With, list):
            With = [With]
        self._with = With
        self._as = As
        self._code_block = None if Do is None else macros.CodeBlock(Do)

    def _resolve_names(self, names, context):
        murano_class = helpers.get_type(context)
        for name in names:
            yield murano_class.namespace_resolver.resolve_name(name)

    def execute(self, context):
        exception = helpers.get_current_exception(context)
        names = (
            None if self._with is None
            else list(self._resolve_names(self._with, context))
        )

        for name in exception.names:
            if self._with is None or name in names:
                if self._code_block:
                    if self._as:
                        wrapped = self._wrap_internal_exception(
                            exception, context, name)
                        context[self._as] = wrapped
                    self._code_block.execute(context)
                return True
        return False

    def _wrap_internal_exception(self, exception, context, name):
        obj = yaql_integration.call_func(context, 'new', 'io.murano.Exception')
        obj.set_property('name', name)
        obj.set_property('message', exception.message)
        obj.set_property('stackTrace', exception.stacktrace)
        obj.set_property('extra', exception.extra)
        obj.set_property('nativeException', exception)
        return obj


class TryBlockMacro(expressions.DslExpression):
    def __init__(self, Try, Catch=None, Finally=None, Else=None):
        self._try_block = macros.CodeBlock(Try)
        self._catch_block = None
        if Catch is not None:
            if not isinstance(Catch, list):
                Catch = [Catch]
            self._catch_block = [CatchBlock(**c) for c in Catch]
        self._finally_block = (
            None if Finally is None
            else macros.CodeBlock(Finally))
        self._else_block = (
            None if Else is None
            else macros.CodeBlock(Else))

    def execute(self, context):
        try:
            self._try_block.execute(context)
        except dsl_exception.MuranoPlException as e:
            caught = False
            if self._catch_block:
                try:
                    context[constants.CTX_CURRENT_EXCEPTION] = e
                    for cb in self._catch_block:
                        if cb.execute(context):
                            caught = True
                            break
                    if not caught:
                        raise
                finally:
                    context[constants.CTX_CURRENT_EXCEPTION] = None
            else:
                raise
        else:
            if self._else_block:
                self._else_block.execute(context)
        finally:
            if self._finally_block:
                self._finally_block.execute(context)


class RethrowMacro(expressions.DslExpression):
    def __init__(self, Rethrow):
        pass

    def execute(self, context):
        exception = context['$?currentException']
        if not exception:
            raise TypeError('Rethrow must be inside Catch')
        raise exception

    def __str__(self):
        return 'Rethrow'


def register():
    expressions.register_macro(ThrowMacro)
    expressions.register_macro(TryBlockMacro)
    expressions.register_macro(RethrowMacro)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_db import options

from murano.db.migration import migration

CONF = cfg.CONF
options.set_defaults(CONF)


class DBCommand(object):

    def upgrade(self, config):
        migration.upgrade(CONF.command.revision, config=config)

    def downgrade(self, config):
        migration.downgrade(CONF.command.revision, config=config)

    def revision(self, config):
        migration.revision(CONF.command.message,
                           CONF.command.autogenerate,
                           config=config)

    def stamp(self, config):
        migration.stamp(CONF.command.revision, config=config)

    def version(self, config):
        print(migration.version())


def add_command_parsers(subparsers):
    command_object = DBCommand()

    parser = subparsers.add_parser('upgrade')
    parser.set_defaults(func=command_object.upgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser('downgrade')
    parser.set_defaults(func=command_object.downgrade)
    parser.add_argument('--revision', nargs='?')

    parser = subparsers.add_parser('stamp')
    parser.add_argument('--revision', nargs='?')
    parser.set_defaults(func=command_object.stamp)

    parser = subparsers.add_parser('revision')
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.set_defaults(func=command_object.revision)

    parser = subparsers.add_parser('version')
    parser.set_defaults(func=command_object.version)


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help='Available commands',
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def main():
    config = migration.get_alembic_config()
    # attach the Murano conf to the Alembic conf
    config.murano_config = CONF

    CONF(project='murano')
    CONF.command.func(config)



# Copyright (c) 2016 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import cfg
import time

from magnumclient import client
from murano.common import auth_utils
from murano.dsl import session_local_storage
from oslo_config import cfg as config

try:
    from magnumclient.common.apiclient import exceptions
except ImportError:
    # NOTE (hongbin): For magnumclient versions before 2.0.0.
    from magnumclient.openstack.common.apiclient import exceptions

CONF = config.CONF


class MagnumClient(object):
    def __init__(self, this, region_name=None):
        self._region_name = region_name
        self._owner = this.find_owner('io.murano.Environment')

    @property
    def _client(self):
        region = self._region_name or (
            None if self._owner is None else self._owner['region'])
        return self._create_magnum_client(region)

    @classmethod
    def init_plugin(cls):
        cls.CONF = cfg.init_config(CONF)

    def _wait_on_status(self, bays, bay_id, wait_status, finish_status):
        while True:
            # sleep 1s to wait bay status changes, this will be useful for
            # the first time we wait for the status, to avoid another 30s
            time.sleep(1)
            status = bays.get(bay_id).status
            if status in wait_status:
                time.sleep(30)
            elif status in finish_status:
                break
            else:
                raise RuntimeError("Unexpected Status: {}".format(status))

    @staticmethod
    @session_local_storage.execution_session_memoize
    def _create_magnum_client(region):
        session = auth_utils.get_token_client_session(conf=CONF)
        params = auth_utils.get_session_client_parameters(
            service_type='container-infra', region=region, conf=CONF,
            session=session)
        return client.Client(**params)

    def create_baymodel(self, args):
        baymodel = self._client.baymodels.create(**args)
        return baymodel.uuid

    def delete_baymodel(self, baymodel_id):
        self._client.baymodels.delete(baymodel_id)

    def get_bay_status(self, bay_id):
        bays = self._client.bays
        bay = bays.get(bay_id)
        return bay.status

    def create_bay(self, args):
        bays = self._client.bays
        bay = bays.create(**args)
        self._wait_on_status(bays, bay.uuid, [None, "CREATE_IN_PROGRESS"],
                             ["CREATE_COMPLETE", "CREATE_FAILED"])
        return bay.uuid

    def delete_bay(self, bay_id):
        bays = self._client.bays
        bays.delete(bay_id)
        try:
            self._wait_on_status(bays, bay_id, ["CREATE_COMPLETE",
                                 "DELETE_IN_PROGRESS", "CREATE_FAILED"],
                                 ["DELETE_COMPLETE"])
        except exceptions.NotFound:
            pass



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Session management functions."""
import threading

from oslo_config import cfg
from oslo_db import exception
from oslo_db import options
from oslo_db.sqlalchemy import session as db_session
from oslo_utils import timeutils

from murano.db import models

CONF = cfg.CONF

options.set_defaults(CONF)

_FACADE = None
_LOCK = threading.Lock()

MAX_LOCK_RETRIES = 10


def _create_facade_lazily():
    global _LOCK, _FACADE

    if _FACADE is None:
        with _LOCK:
            if _FACADE is None:
                _FACADE = db_session.EngineFacade.from_config(CONF,
                                                              sqlite_fk=True)
    return _FACADE


def get_engine():
    facade = _create_facade_lazily()
    return facade.get_engine()


def get_session(**kwargs):
    facade = _create_facade_lazily()
    return facade.get_session(**kwargs)


def get_lock(name, session=None):
    if session is None:
        session = get_session()
        nested = False
    else:
        nested = session.transaction is not None
    return _get_or_create_lock(name, session, nested)


def _get_or_create_lock(name, session, nested, retry=0):
    if nested:
        session.begin_nested()
    else:
        session.begin()
    existing = session.query(models.Lock).get(name)
    if existing is None:
        try:
            # no lock found, creating a new one
            lock = models.Lock(id=name, ts=timeutils.utcnow())
            lock.save(session)
            return session.transaction
            # lock created and acquired
        except exception.DBDuplicateEntry:
            session.rollback()
            if retry >= MAX_LOCK_RETRIES:
                raise
            else:
                # other transaction has created a lock, repeat to acquire
                # via update
                return _get_or_create_lock(name, session, nested, retry + 1)
    else:
        # lock found, acquiring by doing update
        existing.ts = timeutils.utcnow()
        existing.save(session)
        return session.transaction



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
A filter middleware that inspects the requested URI for a version string
and/or Accept headers and attempts to negotiate an API controller to
return
"""

from oslo_log import log as logging

from murano.api import versions
from murano.common.i18n import _LW
from murano.common import wsgi

LOG = logging.getLogger(__name__)


class VersionNegotiationFilter(wsgi.Middleware):
    @classmethod
    def factory(cls, global_conf, **local_conf):
        def filter(app):
            return cls(app)
        return filter

    def __init__(self, app):
        self.versions_app = versions.Controller()
        super(VersionNegotiationFilter, self).__init__(app)

    def process_request(self, req):
        """Try to find a version first in the accept header, then the URL."""
        LOG.debug(("Determining version of request:{method} {path} "
                   "Accept: {accept}").format(method=req.method,
                                              path=req.path,
                                              accept=req.accept))
        LOG.debug("Using url versioning")
        # Remove version in url so it doesn't conflict later
        req_version = self._pop_path_info(req)

        try:
            version = self._match_version_string(req_version)
        except ValueError:
            LOG.warning(_LW("Unknown version. Returning version choices."))
            return self.versions_app

        req.environ['api.version'] = version
        req.path_info = ''.join(('/v', str(version), req.path_info))
        LOG.debug("Matched version: v{version}".format(version=version))
        LOG.debug('new path {path}'.format(path=req.path_info))
        return None

    def _match_version_string(self, subject):
        """Tries to match major and/or minor version

           Given a string, tries to match a major and/or
           minor version number.

           :param subject: The string to check
           :returns version found in the subject
           :raises ValueError if no acceptable version could be found
        """
        if subject in ('v1',):
            major_version = 1
        else:
            raise ValueError()
        return major_version

    def _pop_path_info(self, req):
        """Returns the popped off next segment

           'Pops' off the next segment of PATH_INFO, returns the popped
           segment. Do NOT push it onto SCRIPT_NAME.
        """
        path = req.path_info
        if not path:
            return None
        while path.startswith('/'):
            path = path[1:]
        idx = path.find('/')
        if idx == -1:
            idx = len(path)
        r = path[:idx]
        req.path_info = path[idx:]
        return r
